from collections.abc import Iterator, Sequence
from io import DEFAULT_BUFFER_SIZE, IOBase
from pathlib import Path
from queue import Queue
from threading import Event
from typing import (  # noqa: UP035
    Generic,
    Literal,
    Optional,
    Type,
    TypedDict,
    TypeVar,
    overload,
)

from . import Index
from ._libgit2.ffi import (
    GitCommitC,
    GitObjectC,
    GitProxyOptionsC,
    GitSignatureC,
    _Pointer,
)
from .enums import (
    ApplyLocation,
    BlobFilter,
    BranchType,
    DeltaStatus,
    DiffFind,
    DiffFlag,
    DiffOption,
    DiffStatsFormat,
    FileMode,
    MergeAnalysis,
    MergePreference,
    ObjectType,
    ReferenceFilter,
    ReferenceType,
    ResetMode,
    SortMode,
)
from .filter import Filter

GIT_OBJ_BLOB = Literal[3]
GIT_OBJ_COMMIT = Literal[1]
GIT_OBJ_TAG = Literal[4]
GIT_OBJ_TREE = Literal[2]

LIBGIT2_VER_MAJOR: int
LIBGIT2_VER_MINOR: int
LIBGIT2_VER_REVISION: int
LIBGIT2_VERSION: str
GIT_OID_RAWSZ: int
GIT_OID_HEXSZ: int
GIT_OID_HEX_ZERO: str
GIT_OID_MINPREFIXLEN: int
GIT_OBJECT_ANY: int
GIT_OBJECT_INVALID: int
GIT_OBJECT_COMMIT: int
GIT_OBJECT_TREE: int
GIT_OBJECT_BLOB: int
GIT_OBJECT_TAG: int
GIT_OBJECT_OFS_DELTA: int
GIT_OBJECT_REF_DELTA: int
GIT_FILEMODE_UNREADABLE: int
GIT_FILEMODE_TREE: int
GIT_FILEMODE_BLOB: int
GIT_FILEMODE_BLOB_EXECUTABLE: int
GIT_FILEMODE_LINK: int
GIT_FILEMODE_COMMIT: int
GIT_SORT_NONE: int
GIT_SORT_TOPOLOGICAL: int
GIT_SORT_TIME: int
GIT_SORT_REVERSE: int
GIT_RESET_SOFT: int
GIT_RESET_MIXED: int
GIT_RESET_HARD: int
GIT_REFERENCES_ALL: int
GIT_REFERENCES_BRANCHES: int
GIT_REFERENCES_TAGS: int
GIT_REVSPEC_SINGLE: int
GIT_REVSPEC_RANGE: int
GIT_REVSPEC_MERGE_BASE: int
GIT_BRANCH_LOCAL: int
GIT_BRANCH_REMOTE: int
GIT_BRANCH_ALL: int
GIT_STATUS_CURRENT: int
GIT_STATUS_INDEX_NEW: int
GIT_STATUS_INDEX_MODIFIED: int
GIT_STATUS_INDEX_DELETED: int
GIT_STATUS_INDEX_RENAMED: int
GIT_STATUS_INDEX_TYPECHANGE: int
GIT_STATUS_WT_NEW: int
GIT_STATUS_WT_MODIFIED: int
GIT_STATUS_WT_DELETED: int
GIT_STATUS_WT_TYPECHANGE: int
GIT_STATUS_WT_RENAMED: int
GIT_STATUS_WT_UNREADABLE: int
GIT_STATUS_IGNORED: int
GIT_STATUS_CONFLICTED: int
GIT_CHECKOUT_NONE: int
GIT_CHECKOUT_SAFE: int
GIT_CHECKOUT_FORCE: int
GIT_CHECKOUT_RECREATE_MISSING: int
GIT_CHECKOUT_ALLOW_CONFLICTS: int
GIT_CHECKOUT_REMOVE_UNTRACKED: int
GIT_CHECKOUT_REMOVE_IGNORED: int
GIT_CHECKOUT_UPDATE_ONLY: int
GIT_CHECKOUT_DONT_UPDATE_INDEX: int
GIT_CHECKOUT_NO_REFRESH: int
GIT_CHECKOUT_SKIP_UNMERGED: int
GIT_CHECKOUT_USE_OURS: int
GIT_CHECKOUT_USE_THEIRS: int
GIT_CHECKOUT_DISABLE_PATHSPEC_MATCH: int
GIT_CHECKOUT_SKIP_LOCKED_DIRECTORIES: int
GIT_CHECKOUT_DONT_OVERWRITE_IGNORED: int
GIT_CHECKOUT_CONFLICT_STYLE_MERGE: int
GIT_CHECKOUT_CONFLICT_STYLE_DIFF3: int
GIT_CHECKOUT_DONT_REMOVE_EXISTING: int
GIT_CHECKOUT_DONT_WRITE_INDEX: int
GIT_CHECKOUT_DRY_RUN: int
GIT_CHECKOUT_CONFLICT_STYLE_ZDIFF3: int
GIT_DIFF_NORMAL: int
GIT_DIFF_REVERSE: int
GIT_DIFF_INCLUDE_IGNORED: int
GIT_DIFF_RECURSE_IGNORED_DIRS: int
GIT_DIFF_INCLUDE_UNTRACKED: int
GIT_DIFF_RECURSE_UNTRACKED_DIRS: int
GIT_DIFF_INCLUDE_UNMODIFIED: int
GIT_DIFF_INCLUDE_TYPECHANGE: int
GIT_DIFF_INCLUDE_TYPECHANGE_TREES: int
GIT_DIFF_IGNORE_FILEMODE: int
GIT_DIFF_IGNORE_SUBMODULES: int
GIT_DIFF_IGNORE_CASE: int
GIT_DIFF_INCLUDE_CASECHANGE: int
GIT_DIFF_DISABLE_PATHSPEC_MATCH: int
GIT_DIFF_SKIP_BINARY_CHECK: int
GIT_DIFF_ENABLE_FAST_UNTRACKED_DIRS: int
GIT_DIFF_UPDATE_INDEX: int
GIT_DIFF_INCLUDE_UNREADABLE: int
GIT_DIFF_INCLUDE_UNREADABLE_AS_UNTRACKED: int
GIT_DIFF_INDENT_HEURISTIC: int
GIT_DIFF_IGNORE_BLANK_LINES: int
GIT_DIFF_FORCE_TEXT: int
GIT_DIFF_FORCE_BINARY: int
GIT_DIFF_IGNORE_WHITESPACE: int
GIT_DIFF_IGNORE_WHITESPACE_CHANGE: int
GIT_DIFF_IGNORE_WHITESPACE_EOL: int
GIT_DIFF_SHOW_UNTRACKED_CONTENT: int
GIT_DIFF_SHOW_UNMODIFIED: int
GIT_DIFF_PATIENCE: int
GIT_DIFF_MINIMAL: int
GIT_DIFF_SHOW_BINARY: int
GIT_DIFF_STATS_NONE: int
GIT_DIFF_STATS_FULL: int
GIT_DIFF_STATS_SHORT: int
GIT_DIFF_STATS_NUMBER: int
GIT_DIFF_STATS_INCLUDE_SUMMARY: int
GIT_DIFF_FIND_BY_CONFIG: int
GIT_DIFF_FIND_RENAMES: int
GIT_DIFF_FIND_RENAMES_FROM_REWRITES: int
GIT_DIFF_FIND_COPIES: int
GIT_DIFF_FIND_COPIES_FROM_UNMODIFIED: int
GIT_DIFF_FIND_REWRITES: int
GIT_DIFF_BREAK_REWRITES: int
GIT_DIFF_FIND_AND_BREAK_REWRITES: int
GIT_DIFF_FIND_FOR_UNTRACKED: int
GIT_DIFF_FIND_ALL: int
GIT_DIFF_FIND_IGNORE_LEADING_WHITESPACE: int
GIT_DIFF_FIND_IGNORE_WHITESPACE: int
GIT_DIFF_FIND_DONT_IGNORE_WHITESPACE: int
GIT_DIFF_FIND_EXACT_MATCH_ONLY: int
GIT_DIFF_BREAK_REWRITES_FOR_RENAMES_ONLY: int
GIT_DIFF_FIND_REMOVE_UNMODIFIED: int
GIT_DIFF_FLAG_BINARY: int
GIT_DIFF_FLAG_NOT_BINARY: int
GIT_DIFF_FLAG_VALID_ID: int
GIT_DIFF_FLAG_EXISTS: int
GIT_DIFF_FLAG_VALID_SIZE: int
GIT_DELTA_UNMODIFIED: int
GIT_DELTA_ADDED: int
GIT_DELTA_DELETED: int
GIT_DELTA_MODIFIED: int
GIT_DELTA_RENAMED: int
GIT_DELTA_COPIED: int
GIT_DELTA_IGNORED: int
GIT_DELTA_UNTRACKED: int
GIT_DELTA_TYPECHANGE: int
GIT_DELTA_UNREADABLE: int
GIT_DELTA_CONFLICTED: int
GIT_CONFIG_LEVEL_PROGRAMDATA: int
GIT_CONFIG_LEVEL_SYSTEM: int
GIT_CONFIG_LEVEL_XDG: int
GIT_CONFIG_LEVEL_GLOBAL: int
GIT_CONFIG_LEVEL_LOCAL: int
GIT_CONFIG_LEVEL_WORKTREE: int
GIT_CONFIG_LEVEL_APP: int
GIT_CONFIG_HIGHEST_LEVEL: int
GIT_BLAME_NORMAL: int
GIT_BLAME_TRACK_COPIES_SAME_FILE: int
GIT_BLAME_TRACK_COPIES_SAME_COMMIT_MOVES: int
GIT_BLAME_TRACK_COPIES_SAME_COMMIT_COPIES: int
GIT_BLAME_TRACK_COPIES_ANY_COMMIT_COPIES: int
GIT_BLAME_FIRST_PARENT: int
GIT_BLAME_USE_MAILMAP: int
GIT_BLAME_IGNORE_WHITESPACE: int
GIT_MERGE_ANALYSIS_NONE: int
GIT_MERGE_ANALYSIS_NORMAL: int
GIT_MERGE_ANALYSIS_UP_TO_DATE: int
GIT_MERGE_ANALYSIS_FASTFORWARD: int
GIT_MERGE_ANALYSIS_UNBORN: int
GIT_MERGE_PREFERENCE_NONE: int
GIT_MERGE_PREFERENCE_NO_FASTFORWARD: int
GIT_MERGE_PREFERENCE_FASTFORWARD_ONLY: int
GIT_DESCRIBE_DEFAULT: int
GIT_DESCRIBE_TAGS: int
GIT_DESCRIBE_ALL: int
GIT_STASH_DEFAULT: int
GIT_STASH_KEEP_INDEX: int
GIT_STASH_INCLUDE_UNTRACKED: int
GIT_STASH_INCLUDE_IGNORED: int
GIT_STASH_KEEP_ALL: int
GIT_STASH_APPLY_DEFAULT: int
GIT_STASH_APPLY_REINSTATE_INDEX: int
GIT_APPLY_LOCATION_WORKDIR: int
GIT_APPLY_LOCATION_INDEX: int
GIT_APPLY_LOCATION_BOTH: int
GIT_SUBMODULE_IGNORE_UNSPECIFIED: int
GIT_SUBMODULE_IGNORE_NONE: int
GIT_SUBMODULE_IGNORE_UNTRACKED: int
GIT_SUBMODULE_IGNORE_DIRTY: int
GIT_SUBMODULE_IGNORE_ALL: int
GIT_SUBMODULE_STATUS_IN_HEAD: int
GIT_SUBMODULE_STATUS_IN_INDEX: int
GIT_SUBMODULE_STATUS_IN_CONFIG: int
GIT_SUBMODULE_STATUS_IN_WD: int
GIT_SUBMODULE_STATUS_INDEX_ADDED: int
GIT_SUBMODULE_STATUS_INDEX_DELETED: int
GIT_SUBMODULE_STATUS_INDEX_MODIFIED: int
GIT_SUBMODULE_STATUS_WD_UNINITIALIZED: int
GIT_SUBMODULE_STATUS_WD_ADDED: int
GIT_SUBMODULE_STATUS_WD_DELETED: int
GIT_SUBMODULE_STATUS_WD_MODIFIED: int
GIT_SUBMODULE_STATUS_WD_INDEX_MODIFIED: int
GIT_SUBMODULE_STATUS_WD_WD_MODIFIED: int
GIT_SUBMODULE_STATUS_WD_UNTRACKED: int
GIT_BLOB_FILTER_CHECK_FOR_BINARY: int
GIT_BLOB_FILTER_NO_SYSTEM_ATTRIBUTES: int
GIT_BLOB_FILTER_ATTRIBUTES_FROM_HEAD: int
GIT_BLOB_FILTER_ATTRIBUTES_FROM_COMMIT: int
GIT_FILTER_DRIVER_PRIORITY: int
GIT_FILTER_TO_WORKTREE: int
GIT_FILTER_SMUDGE: int
GIT_FILTER_TO_ODB: int
GIT_FILTER_CLEAN: int
GIT_FILTER_DEFAULT: int
GIT_FILTER_ALLOW_UNSAFE: int
GIT_FILTER_NO_SYSTEM_ATTRIBUTES: int
GIT_FILTER_ATTRIBUTES_FROM_HEAD: int
GIT_FILTER_ATTRIBUTES_FROM_COMMIT: int

T = TypeVar('T')

class _ObjectBase(Generic[T]):
    _pointer: _Pointer[T]
    filemode: FileMode
    id: Oid
    name: str | None
    raw_name: bytes | None
    short_id: str
    type: 'Literal[GIT_OBJ_COMMIT] | Literal[GIT_OBJ_TREE] | Literal[GIT_OBJ_TAG] | Literal[GIT_OBJ_BLOB]'
    type_str: "Literal['commit'] | Literal['tree'] | Literal['tag'] | Literal['blob']"
    author: Signature
    committer: Signature
    tree: Tree
    @overload
    def peel(
        self, target_type: 'Literal[GIT_OBJ_COMMIT, ObjectType.COMMIT] | Type[Commit]'
    ) -> 'Commit': ...
    @overload
    def peel(
        self, target_type: 'Literal[GIT_OBJ_TREE, ObjectType.TREE] | Type[Tree]'
    ) -> 'Tree': ...
    @overload
    def peel(
        self, target_type: 'Literal[GIT_OBJ_TAG, ObjectType.TAG] | Type[Tag]'
    ) -> 'Tag': ...
    @overload
    def peel(
        self, target_type: 'Literal[GIT_OBJ_BLOB, ObjectType.BLOB] | Type[Blob]'
    ) -> 'Blob': ...
    @overload
    def peel(self, target_type: 'None') -> 'Commit|Tree|Tag|Blob': ...
    def read_raw(self) -> bytes: ...
    def __eq__(self, other) -> bool: ...
    def __ge__(self, other) -> bool: ...
    def __gt__(self, other) -> bool: ...
    def __hash__(self) -> int: ...
    def __le__(self, other) -> bool: ...
    def __lt__(self, other) -> bool: ...
    def __ne__(self, other) -> bool: ...

class Object(_ObjectBase[GitObjectC]):
    pass

class Reference:
    name: str
    raw_name: bytes
    raw_shorthand: bytes
    raw_target: Oid | bytes
    shorthand: str
    target: Oid | str
    type: ReferenceType
    def __init__(self, *args) -> None: ...
    def delete(self) -> None: ...
    def log(self) -> Iterator[RefLogEntry]: ...
    @overload
    def peel(self, type: 'Literal[GIT_OBJ_COMMIT] | Type[Commit]') -> 'Commit': ...
    @overload
    def peel(self, type: 'Literal[GIT_OBJ_TREE] | Type[Tree]') -> 'Tree': ...
    @overload
    def peel(self, type: 'Literal[GIT_OBJ_TAG] | Type[Tag]') -> 'Tag': ...
    @overload
    def peel(self, type: 'Literal[GIT_OBJ_BLOB] | Type[Blob]') -> 'Blob': ...
    @overload
    def peel(self, type: 'None' = None) -> 'Commit|Tree|Tag|Blob': ...
    def rename(self, new_name: str) -> None: ...
    def resolve(self) -> Reference: ...
    def set_target(self, target: _OidArg, message: str = ...) -> None: ...
    def __eq__(self, other) -> bool: ...
    def __ge__(self, other) -> bool: ...
    def __gt__(self, other) -> bool: ...
    def __le__(self, other) -> bool: ...
    def __lt__(self, other) -> bool: ...
    def __ne__(self, other) -> bool: ...

class AlreadyExistsError(ValueError): ...

class Blob(Object):
    data: bytes
    is_binary: bool
    size: int
    def diff(
        self,
        blob: Blob = ...,
        flag: int = ...,
        old_as_path: str = ...,
        new_as_path: str = ...,
    ) -> Patch: ...
    def diff_to_buffer(
        self,
        buffer: Optional[bytes | str] = None,
        flag: DiffOption = DiffOption.NORMAL,
        old_as_path: str = ...,
        buffer_as_path: str = ...,
    ) -> Patch: ...
    def _write_to_queue(
        self,
        queue: Queue[bytes],
        ready: Event,
        done: Event,
        chunk_size: int = DEFAULT_BUFFER_SIZE,
        as_path: Optional[str] = None,
        flags: BlobFilter = BlobFilter.CHECK_FOR_BINARY,
        commit_id: Optional[Oid] = None,
    ) -> None: ...
    def __buffer__(self, flags: int) -> memoryview: ...
    def __release_buffer__(self, buffer: memoryview) -> None: ...

class Branch(Reference):
    branch_name: str
    raw_branch_name: bytes
    remote_name: str
    upstream: Branch
    upstream_name: str
    def delete(self) -> None: ...
    def is_checked_out(self) -> bool: ...
    def is_head(self) -> bool: ...
    def rename(self, name: str, force: bool = False) -> 'Branch': ...  # type: ignore[override]

class FetchOptions:
    # incomplete
    depth: int
    proxy_opts: GitProxyOptionsC

class CloneOptions:
    # incomplete
    version: int
    checkout_opts: object
    fetch_opts: FetchOptions
    bare: int
    local: object
    checkout_branch: object
    repository_cb: object
    repository_cb_payload: object
    remote_cb: object
    remote_cb_payload: object

class Commit(_ObjectBase[GitCommitC]):
    _pointer: _Pointer[GitCommitC]
    author: Signature
    commit_time: int
    commit_time_offset: int
    committer: Signature
    gpg_signature: tuple[bytes, bytes]
    message: str
    message_encoding: str
    message_trailers: dict[str, str]
    parent_ids: list[Oid]
    parents: list[Commit]
    raw_message: bytes
    tree: Tree
    tree_id: Oid

class Diff:
    deltas: Iterator[DiffDelta]
    patch: str | None
    patchid: Oid
    stats: DiffStats
    text: str
    def find_similar(
        self,
        flags: DiffFind = DiffFind.FIND_BY_CONFIG,
        rename_threshold: int = 50,
        copy_threshold: int = 50,
        rename_from_rewrite_threshold: int = 50,
        break_rewrite_threshold: int = 60,
        rename_limit: int = 1000,
    ) -> None: ...
    def merge(self, diff: Diff) -> None: ...
    @staticmethod
    def from_c(diff, repo) -> Diff: ...
    @staticmethod
    def parse_diff(git_diff: str | bytes) -> Diff: ...
    def __getitem__(self, index: int) -> Patch | None: ...  # Diff_getitem
    def __iter__(self) -> Iterator[Patch | None]: ...  # -> DiffIter
    def __len__(self) -> int: ...

class DiffDelta:
    flags: DiffFlag
    is_binary: bool
    nfiles: int
    new_file: DiffFile
    old_file: DiffFile
    similarity: int
    status: DeltaStatus
    def status_char(self) -> str: ...

class DiffFile:
    flags: DiffFlag
    id: Oid
    mode: FileMode
    path: str
    raw_path: bytes
    size: int
    @staticmethod
    def from_c(bytes) -> DiffFile: ...

class DiffHunk:
    header: str
    lines: list[DiffLine]
    new_lines: int
    new_start: int
    old_lines: int
    old_start: int

class DiffLine:
    content: str
    content_offset: int
    new_lineno: int
    num_lines: int
    old_lineno: int
    origin: str
    raw_content: bytes

class DiffStats:
    deletions: int
    files_changed: int
    insertions: int
    def format(self, format: DiffStatsFormat, width: int) -> str: ...

class FilterSource:
    # probably incomplete
    repo: object
    pass

class GitError(Exception): ...
class InvalidSpecError(ValueError): ...

class Mailmap:
    def __init__(self, *args) -> None: ...
    def add_entry(
        self,
        real_name: str | None = ...,
        real_email: str | None = ...,
        replace_name: str | None = ...,
        replace_email: str = ...,
    ) -> None: ...
    @staticmethod
    def from_buffer(buffer: str | bytes) -> Mailmap: ...
    @staticmethod
    def from_repository(repository: Repository) -> Mailmap: ...
    def resolve(self, name: str, email: str) -> tuple[str, str]: ...
    def resolve_signature(self, sig: Signature) -> Signature: ...

class Note:
    annotated_id: Oid
    id: Oid
    message: str
    data: bytes
    def remove(
        self, author: Signature, committer: Signature, ref: str = 'refs/notes/commits'
    ) -> None: ...

class Odb:
    backends: Iterator[OdbBackend]
    def __init__(self, *args, **kwargs) -> None: ...
    def add_backend(self, backend: OdbBackend, priority: int) -> None: ...
    def add_disk_alternate(self, path: str | Path) -> None: ...
    def exists(self, oid: _OidArg) -> bool: ...
    def read(self, oid: _OidArg) -> tuple[ObjectType, bytes]: ...
    def read_header(self, oid: _OidArg) -> tuple[ObjectType, int]: ...
    def write(self, type: int, data: bytes | str) -> Oid: ...
    def __contains__(self, other: _OidArg) -> bool: ...
    def __iter__(self) -> Iterator[Oid]: ...  # Odb_as_iter

class OdbBackend:
    def __init__(self, *args, **kwargs) -> None: ...
    def exists(self, oid: _OidArg) -> bool: ...
    def exists_prefix(self, partial_id: _OidArg) -> Oid: ...
    def read(self, oid: _OidArg) -> tuple[int, bytes]: ...
    def read_header(self, oid: _OidArg) -> tuple[int, int]: ...
    def read_prefix(self, oid: _OidArg) -> tuple[int, bytes, Oid]: ...
    def refresh(self) -> None: ...
    def __iter__(self) -> Iterator[Oid]: ...  # OdbBackend_as_iter

class OdbBackendLoose(OdbBackend):
    def __init__(self, *args, **kwargs) -> None: ...

class OdbBackendPack(OdbBackend):
    def __init__(self, *args, **kwargs) -> None: ...

class Oid:
    raw: bytes
    def __init__(self, raw: bytes = ..., hex: str = ...) -> None: ...
    def __eq__(self, other) -> bool: ...
    def __ge__(self, other) -> bool: ...
    def __gt__(self, other) -> bool: ...
    def __hash__(self) -> int: ...
    def __le__(self, other) -> bool: ...
    def __lt__(self, other) -> bool: ...
    def __ne__(self, other) -> bool: ...
    def __bool__(self) -> bool: ...

class Patch:
    data: bytes
    delta: DiffDelta
    hunks: list[DiffHunk]
    line_stats: tuple[int, int, int]  # context, additions, deletions
    text: str | None

    @staticmethod
    def create_from(
        old: Blob | bytes | None,
        new: Blob | bytes | None,
        old_as_path: str = ...,
        new_as_path: str = ...,
        flag: DiffOption = DiffOption.NORMAL,
        context_lines: int = 3,
        interhunk_lines: int = 0,
    ) -> Patch: ...

class RefLogEntry:
    committer: Signature
    message: str
    oid_new: Oid
    oid_old: Oid
    def __init__(self, *args, **kwargs) -> None: ...

class Refdb:
    def __init__(self, *args, **kwargs) -> None: ...
    def compress(self) -> None: ...
    @staticmethod
    def new(repo: Repository) -> Refdb: ...
    @staticmethod
    def open(repo: Repository) -> Refdb: ...
    def set_backend(self, backend: RefdbBackend) -> None: ...

class RefdbBackend:
    def __init__(self, *args, **kwargs) -> None: ...
    def compress(self) -> None: ...
    def delete(
        self, ref_name: str, old_id: _OidArg, old_target: str | None
    ) -> None: ...
    def ensure_log(self, ref_name: str) -> bool: ...
    def exists(self, refname: str) -> bool: ...
    def has_log(self, ref_name: str) -> bool: ...
    def lookup(self, refname: str) -> Reference: ...
    def rename(
        self, old_name: str, new_name: str, force: bool, who: Signature, message: str
    ) -> Reference: ...
    def write(
        self,
        ref: Reference,
        force: bool,
        who: Signature,
        message: str,
        old: None | _OidArg,
        old_target: None | str,
    ) -> None: ...
    def __iter__(self) -> Iterator[Reference]: ...

class RefdbFsBackend(RefdbBackend):
    def __init__(self, *args, **kwargs) -> None: ...

_Proxy = None | Literal[True] | str

class _StrArray:
    # incomplete
    count: int

class PushOptions:
    version: int
    pb_parallelism: int
    callbacks: object  # TODO
    proxy_opts: GitProxyOptionsC
    follow_redirects: object  # TODO
    custom_headers: _StrArray
    remote_push_options: _StrArray

class _LsRemotesDict(TypedDict):
    local: bool
    loid: Oid | None
    name: str | None
    symref_target: str | None
    oid: Oid

class Repository:
    def TreeBuilder(self, src: Tree | _OidArg = ...) -> TreeBuilder: ...
    def _disown(self, *args, **kwargs) -> None: ...
    def add_worktree(
        self, name: str, path: str | Path, ref: Reference = ...
    ) -> Worktree: ...
    def applies(
        self,
        diff: Diff,
        location: ApplyLocation = ApplyLocation.INDEX,
        raise_error: bool = False,
    ) -> bool: ...
    def apply(
        self, diff: Diff, location: ApplyLocation = ApplyLocation.WORKDIR
    ) -> None: ...
    def cherrypick(self, id: _OidArg) -> None: ...
    def compress_references(self) -> None: ...
    def create_blob(self, data: str | bytes) -> Oid: ...
    def create_blob_fromdisk(self, path: str) -> Oid: ...
    def create_blob_fromiobase(self, iobase: IOBase) -> Oid: ...
    def create_blob_fromworkdir(self, path: str | Path) -> Oid: ...
    def create_branch(self, name: str, commit: Commit, force=False) -> Branch: ...
    def create_commit(
        self,
        reference_name: Optional[str],
        author: Signature,
        committer: Signature,
        message: str | bytes,
        tree: _OidArg,
        parents: Sequence[_OidArg],
        encoding: str = ...,
    ) -> Oid: ...
    def create_commit_string(
        self,
        author: Signature,
        committer: Signature,
        message: str | bytes,
        tree: _OidArg,
        parents: list[_OidArg],
        encoding: str = ...,
    ) -> Oid: ...
    def create_commit_with_signature(
        self, content: str, signature: str, signature_field: Optional[str] = None
    ) -> Oid: ...
    def create_note(
        self,
        message: str,
        author: Signature,
        committer: Signature,
        annotated_id: str,
        ref: str = 'refs/notes/commits',
        force: bool = False,
    ) -> Oid: ...
    def create_reference_direct(
        self, name: str, target: _OidArg, force: bool, message: Optional[str] = None
    ) -> Reference: ...
    def create_reference_symbolic(
        self, name: str, target: str, force: bool, message: Optional[str] = None
    ) -> Reference: ...
    def create_tag(
        self, name: str, oid: _OidArg, type: ObjectType, tagger: Signature, message: str
    ) -> Oid: ...
    def descendant_of(self, oid1: _OidArg, oid2: _OidArg) -> bool: ...
    def expand_id(self, hex: str) -> Oid: ...
    def free(self) -> None: ...
    def git_object_lookup_prefix(self, oid: _OidArg) -> Object: ...
    def list_worktrees(self) -> list[str]: ...
    def listall_branches(self, flag: BranchType = BranchType.LOCAL) -> list[str]: ...
    def listall_mergeheads(self) -> list[Oid]: ...
    def listall_stashes(self) -> list[Stash]: ...
    def listall_submodules(self) -> list[str]: ...
    def lookup_branch(
        self, branch_name: str | bytes, branch_type: BranchType = BranchType.LOCAL
    ) -> Branch: ...
    def lookup_note(
        self, annotated_id: str, ref: str = 'refs/notes/commits'
    ) -> Note: ...
    def lookup_reference(self, name: str) -> Reference: ...
    def lookup_reference_dwim(self, name: str) -> Reference: ...
    def lookup_worktree(self, name: str) -> Worktree: ...
    def merge_analysis(
        self, their_head: _OidArg, our_ref: str = 'HEAD'
    ) -> tuple[MergeAnalysis, MergePreference]: ...
    def merge_base(self, oid1: _OidArg, oid2: _OidArg) -> Oid: ...
    def merge_base_many(self, oids: list[_OidArg]) -> Oid: ...
    def merge_base_octopus(self, oids: list[_OidArg]) -> Oid: ...
    def notes(self) -> Iterator[Note]: ...
    def path_is_ignored(self, path: str) -> bool: ...
    def raw_listall_branches(
        self, flag: BranchType = BranchType.LOCAL
    ) -> list[bytes]: ...
    def raw_listall_references(self) -> list[bytes]: ...
    def references_iterator_init(self) -> Iterator[Reference]: ...
    def references_iterator_next(
        self,
        iter: Iterator[T],
        references_return_type: ReferenceFilter = ReferenceFilter.ALL,
    ) -> Reference: ...
    def reset(self, oid: _OidArg, reset_type: ResetMode) -> None: ...
    def revparse(self, revspec: str) -> RevSpec: ...
    def revparse_ext(self, revision: str) -> tuple[Object, Reference]: ...
    def revparse_single(self, revision: str) -> Object: ...
    def set_odb(self, odb: Odb) -> None: ...
    def set_refdb(self, refdb: Refdb) -> None: ...
    def status(
        self, untracked_files: str = 'all', ignored: bool = False
    ) -> dict[str, int]: ...
    def status_file(self, path: str) -> int: ...
    def walk(
        self, oid: _OidArg | None, sort_mode: SortMode = SortMode.NONE
    ) -> Walker: ...

class RevSpec:
    flags: int
    from_object: Object
    to_object: Object

class Signature:
    _encoding: str | None
    _pointer: _Pointer[GitSignatureC]
    email: str
    name: str
    offset: int
    raw_email: bytes
    raw_name: bytes
    time: int
    def __init__(
        self,
        name: str | bytes,
        email: str,
        time: int = -1,
        offset: int = 0,
        encoding: Optional[str] = None,
    ) -> None: ...
    def __eq__(self, other) -> bool: ...
    def __ge__(self, other) -> bool: ...
    def __gt__(self, other) -> bool: ...
    def __le__(self, other) -> bool: ...
    def __lt__(self, other) -> bool: ...
    def __ne__(self, other) -> bool: ...

class Stash:
    commit_id: Oid
    message: str
    raw_message: bytes
    def __eq__(self, other) -> bool: ...
    def __ge__(self, other) -> bool: ...
    def __gt__(self, other) -> bool: ...
    def __le__(self, other) -> bool: ...
    def __lt__(self, other) -> bool: ...
    def __ne__(self, other) -> bool: ...

class Tag(Object):
    message: str
    name: str
    raw_message: bytes
    raw_name: bytes
    tagger: Signature
    target: Oid
    def get_object(self) -> Object: ...

class Tree(Object):
    def diff_to_index(
        self,
        index: Index,
        flags: DiffOption = DiffOption.NORMAL,
        context_lines: int = 3,
        interhunk_lines: int = 0,
    ) -> Diff: ...
    def diff_to_tree(
        self,
        tree: Tree = ...,
        flags: DiffOption = DiffOption.NORMAL,
        context_lines: int = 3,
        interhunk_lines: int = 3,
        swap: bool = False,
    ) -> Diff: ...
    def diff_to_workdir(
        self,
        flags: DiffOption = DiffOption.NORMAL,
        context_lines: int = 3,
        interhunk_lines: int = 0,
    ) -> Diff: ...
    def __contains__(self, other: str) -> bool: ...  # Tree_contains
    def __getitem__(self, index: str | int) -> Tree | Blob: ...  # Tree_subscript
    def __iter__(self) -> Iterator[Object]: ...
    def __len__(self) -> int: ...  # Tree_len
    def __rtruediv__(self, other: str) -> Tree | Blob: ...
    def __truediv__(self, other: str) -> Tree | Blob: ...  # Tree_divide

class TreeBuilder:
    def clear(self) -> None: ...
    def get(self, name: str) -> Object: ...
    def insert(self, name: str, oid: _OidArg, attr: int) -> None: ...
    def remove(self, name: str) -> None: ...
    def write(self) -> Oid: ...
    def __len__(self) -> int: ...

class Walker:
    def hide(self, oid: _OidArg) -> None: ...
    def push(self, oid: _OidArg) -> None: ...
    def reset(self) -> None: ...
    def simplify_first_parent(self) -> None: ...
    def sort(self, mode: SortMode) -> None: ...
    def __iter__(self) -> Iterator[Commit]: ...  # Walker: ...
    def __next__(self) -> Commit: ...

class Worktree:
    is_prunable: bool
    name: str
    path: str
    def prune(self, force=False) -> None: ...

def discover_repository(
    path: str | Path, across_fs: bool = False, ceiling_dirs: str = ...
) -> str | None: ...
def hash(data: bytes | str) -> Oid: ...
def hashfile(path: str) -> Oid: ...
def init_file_backend(path: str, flags: int = 0) -> object: ...
def reference_is_valid_name(refname: str) -> bool: ...
def tree_entry_cmp(a: Object, b: Object) -> int: ...
def _cache_enums() -> None: ...
def filter_register(name: str, filter: type[Filter]) -> None: ...

_OidArg = str | Oid



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

from __future__ import annotations

import warnings
from collections.abc import Generator, Iterator
from typing import TYPE_CHECKING, Any, Literal

# Import from pygit2
from pygit2 import RemoteCallbacks

from . import utils
from ._pygit2 import Oid
from .callbacks import (
    git_fetch_options,
    git_proxy_options,
    git_push_options,
    git_remote_callbacks,
)
from .enums import FetchPrune
from .errors import check_error
from .ffi import C, ffi
from .refspec import Refspec
from .utils import StrArray, maybe_string, strarray_to_strings, to_bytes

# Need BaseRepository for type hints, but don't let it cause a circular dependency
if TYPE_CHECKING:
    from ._libgit2.ffi import GitRemoteC, char_pointer
    from .repository import BaseRepository


class RemoteHead:
    """
    Description of a reference advertised by a remote server,
    given out on `Remote.list_heads` calls.
    """

    local: bool
    """Available locally"""

    oid: Oid

    loid: Oid

    name: str | None

    symref_target: str | None
    """
    If the server sent a symref mapping for this ref, this will
    point to the target. 
    """

    def __init__(self, c_struct: Any) -> None:
        self.local = bool(c_struct.local)
        self.oid = Oid(raw=bytes(ffi.buffer(c_struct.oid.id)[:]))
        self.loid = Oid(raw=bytes(ffi.buffer(c_struct.loid.id)[:]))
        self.name = maybe_string(c_struct.name)
        self.symref_target = maybe_string(c_struct.symref_target)


class PushUpdate:
    """
    Represents an update which will be performed on the remote during push.
    """

    src_refname: str
    """The source name of the reference"""

    dst_refname: str
    """The name of the reference to update on the server"""

    src: Oid
    """The current target of the reference"""

    dst: Oid
    """The new target for the reference"""

    def __init__(self, c_struct: Any) -> None:
        src_refname = maybe_string(c_struct.src_refname)
        dst_refname = maybe_string(c_struct.dst_refname)
        assert src_refname is not None, 'libgit2 returned null src_refname'
        assert dst_refname is not None, 'libgit2 returned null dst_refname'
        self.src_refname = src_refname
        self.dst_refname = dst_refname
        self.src = Oid(raw=bytes(ffi.buffer(c_struct.src.id)[:]))
        self.dst = Oid(raw=bytes(ffi.buffer(c_struct.dst.id)[:]))


class TransferProgress:
    """Progress downloading and indexing data during a fetch."""

    total_objects: int
    indexed_objects: int
    received_objects: int
    local_objects: int
    total_deltas: int
    indexed_deltas: int
    received_bytes: int

    def __init__(self, tp: Any) -> None:
        self.total_objects = tp.total_objects
        """Total number of objects to download"""

        self.indexed_objects = tp.indexed_objects
        """Objects which have been indexed"""

        self.received_objects = tp.received_objects
        """Objects which have been received up to now"""

        self.local_objects = tp.local_objects
        """Local objects which were used to fix the thin pack"""

        self.total_deltas = tp.total_deltas
        """Total number of deltas in the pack"""

        self.indexed_deltas = tp.indexed_deltas
        """Deltas which have been indexed"""

        self.received_bytes = tp.received_bytes
        """"Number of bytes received up to now"""


class Remote:
    def __init__(self, repo: BaseRepository, ptr: 'GitRemoteC') -> None:
        """The constructor is for internal use only."""
        self._repo = repo
        self._remote = ptr
        self._stored_exception = None

    def __del__(self) -> None:
        C.git_remote_free(self._remote)

    @property
    def name(self) -> str | None:
        """Name of the remote"""

        return maybe_string(C.git_remote_name(self._remote))

    @property
    def url(self) -> str | None:
        """Url of the remote"""

        return maybe_string(C.git_remote_url(self._remote))

    @property
    def push_url(self) -> str | None:
        """Push url of the remote"""

        return maybe_string(C.git_remote_pushurl(self._remote))

    def connect(
        self,
        callbacks: RemoteCallbacks | None = None,
        direction: int = C.GIT_DIRECTION_FETCH,
        proxy: None | bool | str = None,
    ) -> None:
        """Connect to the remote.

        Parameters:

        proxy : None or True or str
            Proxy configuration. Can be one of:

            * `None` (the default) to disable proxy usage
            * `True` to enable automatic proxy detection
            * an url to a proxy (`http://proxy.example.org:3128/`)
        """
        with git_proxy_options(self, proxy=proxy) as proxy_opts:
            with git_remote_callbacks(callbacks) as payload:
                err = C.git_remote_connect(
                    self._remote,
                    direction,
                    payload.remote_callbacks,
                    proxy_opts,
                    ffi.NULL,
                )
                payload.check_error(err)

    def fetch(
        self,
        refspecs: list[str] | None = None,
        message: str | None = None,
        callbacks: RemoteCallbacks | None = None,
        prune: FetchPrune = FetchPrune.UNSPECIFIED,
        proxy: None | Literal[True] | str = None,
        depth: int = 0,
    ) -> TransferProgress:
        """Perform a fetch against this remote. Returns a <TransferProgress>
        object.

        Parameters:

        prune : enums.FetchPrune
            * `UNSPECIFIED`: use the configuration from the repository.
            * `PRUNE`: remove any remote branch in the local repository
               that does not exist in the remote.
            * `NO_PRUNE`: always keep the remote branches

        proxy : None or True or str
            Proxy configuration. Can be one of:

            * `None` (the default) to disable proxy usage
            * `True` to enable automatic proxy detection
            * an url to a proxy (`http://proxy.example.org:3128/`)

        depth : int
            Number of commits from the tip of each remote branch history to fetch.

            If non-zero, the number of commits from the tip of each remote
            branch history to fetch. If zero, all history is fetched.
            The default is 0 (all history is fetched).
        """
        with git_fetch_options(callbacks) as payload:
            opts = payload.fetch_options
            opts.prune = prune
            opts.depth = depth
            with git_proxy_options(self, payload.fetch_options.proxy_opts, proxy):
                with StrArray(refspecs) as arr:
                    err = C.git_remote_fetch(
                        self._remote, arr.ptr, opts, to_bytes(message)
                    )
                    payload.check_error(err)

        return TransferProgress(C.git_remote_stats(self._remote))

    def list_heads(
        self,
        callbacks: RemoteCallbacks | None = None,
        proxy: str | None | bool = None,
        connect: bool = True,
    ) -> list[RemoteHead]:
        """
        Get the list of references with which the server responds to a new
        connection.

        Parameters:

        callbacks : Passed to connect()

        proxy : Passed to connect()

        connect : Whether to connect to the remote first. You can pass False
        if the remote has already connected. The list remains available after
        disconnecting as long as a new connection is not initiated.
        """

        if connect:
            self.connect(callbacks=callbacks, proxy=proxy)

        refs_ptr = ffi.new('git_remote_head ***')
        size_ptr = ffi.new('size_t *')

        err = C.git_remote_ls(refs_ptr, size_ptr, self._remote)
        check_error(err)

        num_refs = int(size_ptr[0])
        results = [RemoteHead(refs_ptr[0][i]) for i in range(num_refs)]

        return results

    def ls_remotes(
        self,
        callbacks: RemoteCallbacks | None = None,
        proxy: str | None | bool = None,
        connect: bool = True,
    ) -> list[dict[str, Any]]:
        """
        Deprecated interface to list_heads
        """
        warnings.warn('Use list_heads', DeprecationWarning)

        heads = self.list_heads(callbacks, proxy, connect)

        return [
            {
                'local': h.local,
                'oid': h.oid,
                'loid': h.loid if h.local else None,
                'name': h.name,
                'symref_target': h.symref_target,
            }
            for h in heads
        ]

    def prune(self, callbacks: RemoteCallbacks | None = None) -> None:
        """Perform a prune against this remote."""
        with git_remote_callbacks(callbacks) as payload:
            err = C.git_remote_prune(self._remote, payload.remote_callbacks)
            payload.check_error(err)

    @property
    def refspec_count(self) -> int:
        """Total number of refspecs in this remote"""

        return C.git_remote_refspec_count(self._remote)

    def get_refspec(self, n: int) -> Refspec:
        """Return the <Refspec> object at the given position."""
        spec = C.git_remote_get_refspec(self._remote, n)
        return Refspec(self, spec)

    @property
    def fetch_refspecs(self) -> list[str]:
        """Refspecs that will be used for fetching"""

        specs = ffi.new('git_strarray *')
        err = C.git_remote_get_fetch_refspecs(specs, self._remote)
        check_error(err)
        return strarray_to_strings(specs)

    @property
    def push_refspecs(self) -> list[str]:
        """Refspecs that will be used for pushing"""

        specs = ffi.new('git_strarray *')
        err = C.git_remote_get_push_refspecs(specs, self._remote)
        check_error(err)
        return strarray_to_strings(specs)

    def push(
        self,
        specs: list[str],
        callbacks: RemoteCallbacks | None = None,
        proxy: None | bool | str = None,
        push_options: None | list[str] = None,
        threads: int = 1,
    ) -> None:
        """
        Push the given refspec to the remote. Raises ``GitError`` on protocol
        error or unpack failure.

        When the remote has a githook installed, that denies the reference this
        function will return successfully. Thus it is strongly recommended to
        install a callback, that implements
        :py:meth:`RemoteCallbacks.push_update_reference` and check the passed
        parameters for successful operations.

        Parameters:

        specs : [str]
            Push refspecs to use.

        callbacks :

        proxy : None or True or str
            Proxy configuration. Can be one of:

            * `None` (the default) to disable proxy usage
            * `True` to enable automatic proxy detection
            * an url to a proxy (`http://proxy.example.org:3128/`)

        push_options : [str]
            Push options to send to the server, which passes them to the
            pre-receive as well as the post-receive hook.

        threads : int
            If the transport being used to push to the remote requires the
            creation of a pack file, this controls the number of worker threads
            used by the packbuilder when creating that pack file to be sent to
            the remote.

            If set to 0, the packbuilder will auto-detect the number of threads
            to create. The default value is 1.
        """
        with git_push_options(callbacks) as payload:
            opts = payload.push_options
            opts.pb_parallelism = threads
            with git_proxy_options(self, payload.push_options.proxy_opts, proxy):
                with StrArray(specs) as refspecs, StrArray(push_options) as pushopts:
                    pushopts.assign_to(opts.remote_push_options)
                    err = C.git_remote_push(self._remote, refspecs.ptr, opts)
                    payload.check_error(err)


class RemoteCollection:
    """Collection of configured remotes

    You can use this class to look up and manage the remotes configured
    in a repository.  You can access repositories using index
    access. E.g. to look up the "origin" remote, you can use

    >>> repo.remotes["origin"]
    """

    def __init__(self, repo: BaseRepository) -> None:
        self._repo = repo

    def __len__(self) -> int:
        with utils.new_git_strarray() as names:
            err = C.git_remote_list(names, self._repo._repo)
            check_error(err)
            return names.count

    def __iter__(self) -> Iterator[Remote]:
        cremote = ffi.new('git_remote **')
        for name in self._ffi_names():
            err = C.git_remote_lookup(cremote, self._repo._repo, name)
            check_error(err)

            yield Remote(self._repo, cremote[0])

    def __getitem__(self, name: str | int) -> Remote:
        if isinstance(name, int):
            return list(self)[name]

        cremote = ffi.new('git_remote **')
        err = C.git_remote_lookup(cremote, self._repo._repo, to_bytes(name))
        check_error(err)

        return Remote(self._repo, cremote[0])

    def _ffi_names(self) -> Generator['char_pointer', None, None]:
        with utils.new_git_strarray() as names:
            err = C.git_remote_list(names, self._repo._repo)
            check_error(err)
            for i in range(names.count):
                yield names.strings[i]  # type: ignore[index]

    def names(self) -> Generator[str | None, None, None]:
        """An iterator over the names of the available remotes."""
        for name in self._ffi_names():
            yield maybe_string(name)

    def create(self, name: str, url: str, fetch: str | None = None) -> Remote:
        """Create a new remote with the given name and url. Returns a <Remote>
        object.

        If 'fetch' is provided, this fetch refspec will be used instead of the
        default.
        """
        cremote = ffi.new('git_remote **')

        name_bytes = to_bytes(name)
        url_bytes = to_bytes(url)
        if fetch:
            fetch_bytes = to_bytes(fetch)
            err = C.git_remote_create_with_fetchspec(
                cremote, self._repo._repo, name_bytes, url_bytes, fetch_bytes
            )
        else:
            err = C.git_remote_create(cremote, self._repo._repo, name_bytes, url_bytes)

        check_error(err)

        return Remote(self._repo, cremote[0])

    def create_anonymous(self, url: str) -> Remote:
        """Create a new anonymous (in-memory only) remote with the given URL.
        Returns a <Remote> object.
        """
        cremote = ffi.new('git_remote **')
        url_bytes = to_bytes(url)
        err = C.git_remote_create_anonymous(cremote, self._repo._repo, url_bytes)
        check_error(err)
        return Remote(self._repo, cremote[0])

    def rename(self, name: str, new_name: str) -> list[str]:
        """Rename a remote in the configuration. The refspecs in standard
        format will be renamed.

        Returns a list of fetch refspecs (list of strings) which were not in
        the standard format and thus could not be remapped.
        """

        if not name:
            raise ValueError('Current remote name must be a non-empty string')

        if not new_name:
            raise ValueError('New remote name must be a non-empty string')

        problems = ffi.new('git_strarray *')
        err = C.git_remote_rename(
            problems, self._repo._repo, to_bytes(name), to_bytes(new_name)
        )
        check_error(err)
        return strarray_to_strings(problems)

    def delete(self, name: str) -> None:
        """Remove a remote from the configuration

        All remote-tracking branches and configuration settings for the remote will be removed.
        """
        err = C.git_remote_delete(self._repo._repo, to_bytes(name))
        check_error(err)

    def set_url(self, name: str, url: str) -> None:
        """Set the URL for a remote"""
        err = C.git_remote_set_url(self._repo._repo, to_bytes(name), to_bytes(url))
        check_error(err)

    def set_push_url(self, name: str, url: str) -> None:
        """Set the push-URL for a remote"""
        err = C.git_remote_set_pushurl(self._repo._repo, to_bytes(name), to_bytes(url))
        check_error(err)

    def add_fetch(self, name: str, refspec: str) -> None:
        """Add a fetch refspec (str) to the remote"""

        err = C.git_remote_add_fetch(
            self._repo._repo, to_bytes(name), to_bytes(refspec)
        )
        check_error(err)

    def add_push(self, name: str, refspec: str) -> None:
        """Add a push refspec (str) to the remote"""

        err = C.git_remote_add_push(self._repo._repo, to_bytes(name), to_bytes(refspec))
        check_error(err)



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

from collections.abc import Iterator
from typing import TYPE_CHECKING

from ._pygit2 import Oid, Repository, Signature

# Import from pygit2
from .ffi import C, ffi
from .utils import GenericIterator

if TYPE_CHECKING:
    from ._libgit2.ffi import GitBlameC, GitHunkC, GitSignatureC


def wrap_signature(csig: 'GitSignatureC') -> None | Signature:
    if not csig:
        return None

    return Signature(
        ffi.string(csig.name).decode('utf-8'),
        ffi.string(csig.email).decode('utf-8'),
        csig.when.time,
        csig.when.offset,
        'utf-8',
    )


class BlameHunk:
    _blame: 'Blame'
    _hunk: 'GitHunkC'

    @classmethod
    def _from_c(cls, blame: 'Blame', ptr: 'GitHunkC') -> 'BlameHunk':
        hunk = cls.__new__(cls)
        hunk._blame = blame
        hunk._hunk = ptr
        return hunk

    @property
    def lines_in_hunk(self) -> int:
        """Number of lines"""
        return self._hunk.lines_in_hunk

    @property
    def boundary(self) -> bool:
        """Tracked to a boundary commit"""
        # Casting directly to bool via cffi does not seem to work
        return int(ffi.cast('int', self._hunk.boundary)) != 0

    @property
    def final_start_line_number(self) -> int:
        """Final start line number"""
        return self._hunk.final_start_line_number

    @property
    def final_committer(self) -> None | Signature:
        """Final committer"""
        return wrap_signature(self._hunk.final_signature)

    @property
    def final_commit_id(self) -> Oid:
        return Oid(
            raw=bytes(ffi.buffer(ffi.addressof(self._hunk, 'final_commit_id'))[:])
        )

    @property
    def orig_start_line_number(self) -> int:
        """Origin start line number"""
        return self._hunk.orig_start_line_number

    @property
    def orig_committer(self) -> None | Signature:
        """Original committer"""
        return wrap_signature(self._hunk.orig_signature)

    @property
    def orig_commit_id(self) -> Oid:
        return Oid(
            raw=bytes(ffi.buffer(ffi.addressof(self._hunk, 'orig_commit_id'))[:])
        )

    @property
    def orig_path(self) -> None | str:
        """Original path"""
        path = self._hunk.orig_path
        if not path:
            return None

        return ffi.string(path).decode('utf-8')


class Blame:
    _repo: Repository
    _blame: 'GitBlameC'

    @classmethod
    def _from_c(cls, repo: Repository, ptr: 'GitBlameC') -> 'Blame':
        blame = cls.__new__(cls)
        blame._repo = repo
        blame._blame = ptr
        return blame

    def __del__(self) -> None:
        C.git_blame_free(self._blame)

    def __len__(self) -> int:
        return C.git_blame_get_hunk_count(self._blame)

    def __getitem__(self, index: int) -> BlameHunk:
        chunk = C.git_blame_get_hunk_byindex(self._blame, index)
        if not chunk:
            raise IndexError

        return BlameHunk._from_c(self, chunk)

    def for_line(self, line_no: int) -> BlameHunk:
        """
        Returns the <BlameHunk> object for a given line given its number in the
        current Blame.

        Parameters:

        line_no
            Line number, starts at 1.
        """
        if line_no < 0:
            raise IndexError

        chunk = C.git_blame_get_hunk_byline(self._blame, line_no)
        if not chunk:
            raise IndexError

        return BlameHunk._from_c(self, chunk)

    def __iter__(self) -> Iterator[BlameHunk]:
        return GenericIterator(self)



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

from __future__ import annotations

from typing import TYPE_CHECKING

from .enums import CredentialType

if TYPE_CHECKING:
    from pathlib import Path


class Username:
    """Username credentials

    This is an object suitable for passing to a remote's credentials
    callback and for returning from said callback.
    """

    def __init__(self, username: str):
        self._username = username

    @property
    def credential_type(self) -> CredentialType:
        return CredentialType.USERNAME

    @property
    def credential_tuple(self) -> tuple[str]:
        return (self._username,)

    def __call__(
        self, _url: str, _username: str | None, _allowed: CredentialType
    ) -> Username:
        return self


class UserPass:
    """Username/Password credentials

    This is an object suitable for passing to a remote's credentials
    callback and for returning from said callback.
    """

    def __init__(self, username: str, password: str):
        self._username = username
        self._password = password

    @property
    def credential_type(self) -> CredentialType:
        return CredentialType.USERPASS_PLAINTEXT

    @property
    def credential_tuple(self) -> tuple[str, str]:
        return (self._username, self._password)

    def __call__(
        self, _url: str, _username: str | None, _allowed: CredentialType
    ) -> UserPass:
        return self


class Keypair:
    """
    SSH key pair credentials.

    This is an object suitable for passing to a remote's credentials
    callback and for returning from said callback.

    Parameters:

    username : str
        The username being used to authenticate with the remote server.

    pubkey : str
        The path to the user's public key file.

    privkey : str
        The path to the user's private key file.

    passphrase : str
        The password used to decrypt the private key file, or empty string if
        no passphrase is required.
    """

    def __init__(
        self,
        username: str,
        pubkey: str | Path | None,
        privkey: str | Path | None,
        passphrase: str | None,
    ):
        self._username = username
        self._pubkey = pubkey
        self._privkey = privkey
        self._passphrase = passphrase

    @property
    def credential_type(self) -> CredentialType:
        return CredentialType.SSH_KEY

    @property
    def credential_tuple(
        self,
    ) -> tuple[str, str | Path | None, str | Path | None, str | None]:
        return (self._username, self._pubkey, self._privkey, self._passphrase)

    def __call__(
        self, _url: str, _username: str | None, _allowed: CredentialType
    ) -> Keypair:
        return self


class KeypairFromAgent(Keypair):
    def __init__(self, username: str):
        super().__init__(username, None, None, None)


class KeypairFromMemory(Keypair):
    @property
    def credential_type(self) -> CredentialType:
        return CredentialType.SSH_MEMORY



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

from __future__ import annotations

from collections.abc import Iterable, Iterator
from pathlib import Path
from typing import TYPE_CHECKING, Optional

from ._pygit2 import Oid
from .callbacks import RemoteCallbacks, git_fetch_options
from .enums import SubmoduleIgnore, SubmoduleStatus
from .errors import check_error
from .ffi import C, ffi
from .utils import maybe_string, to_bytes

# Need BaseRepository for type hints, but don't let it cause a circular dependency
if TYPE_CHECKING:
    from pygit2 import Repository
    from pygit2._libgit2.ffi import GitSubmoduleC

    from .repository import BaseRepository


class Submodule:
    _repo: BaseRepository
    _subm: 'GitSubmoduleC'

    @classmethod
    def _from_c(cls, repo: BaseRepository, cptr: 'GitSubmoduleC') -> 'Submodule':
        subm = cls.__new__(cls)

        subm._repo = repo
        subm._subm = cptr

        return subm

    def __del__(self) -> None:
        C.git_submodule_free(self._subm)

    def open(self) -> Repository:
        """Open the repository for a submodule."""
        crepo = ffi.new('git_repository **')
        err = C.git_submodule_open(crepo, self._subm)
        check_error(err)

        return self._repo._from_c(crepo[0], True)  # type: ignore[attr-defined]

    def init(self, overwrite: bool = False) -> None:
        """
        Just like "git submodule init", this copies information about the submodule
        into ".git/config".

        Parameters:

        overwrite
            By default, existing submodule entries will not be overwritten,
            but setting this to True forces them to be updated.
        """
        err = C.git_submodule_init(self._subm, int(overwrite))
        check_error(err)

    def update(
        self,
        init: bool = False,
        callbacks: Optional[RemoteCallbacks] = None,
        depth: int = 0,
    ) -> None:
        """
        Update a submodule. This will clone a missing submodule and checkout
        the subrepository to the commit specified in the index of the
        containing repository. If the submodule repository doesn't contain the
        target commit (e.g. because fetchRecurseSubmodules isn't set), then the
        submodule is fetched using the fetch options supplied in options.

        Parameters:

        init
            If the submodule is not initialized, setting this flag to True will
            initialize the submodule before updating. Otherwise, this will raise
            an error if attempting to update an uninitialized repository.

        callbacks
            Optional RemoteCallbacks to clone or fetch the submodule.

        depth
            Number of commits to fetch.
            The default is 0 (full commit history).
        """

        opts = ffi.new('git_submodule_update_options *')
        C.git_submodule_update_options_init(
            opts, C.GIT_SUBMODULE_UPDATE_OPTIONS_VERSION
        )
        opts.fetch_opts.depth = depth

        with git_fetch_options(callbacks, opts=opts.fetch_opts) as payload:
            err = C.git_submodule_update(self._subm, int(init), opts)
            payload.check_error(err)

    def reload(self, force: bool = False) -> None:
        """
        Reread submodule info from config, index, and HEAD.

        Call this to reread cached submodule information for this submodule if
        you have reason to believe that it has changed.

        Parameters:

        force
            Force reload even if the data doesn't seem out of date
        """
        err = C.git_submodule_reload(self._subm, int(force))
        check_error(err)

    @property
    def name(self):
        """Name of the submodule."""
        name = C.git_submodule_name(self._subm)
        return ffi.string(name).decode('utf-8')

    @property
    def path(self):
        """Path of the submodule."""
        path = C.git_submodule_path(self._subm)
        return ffi.string(path).decode('utf-8')

    @property
    def url(self) -> str | None:
        """URL of the submodule."""
        url = C.git_submodule_url(self._subm)
        return maybe_string(url)

    @url.setter
    def url(self, url: str) -> None:
        crepo = self._repo._repo
        cname = ffi.new('char[]', to_bytes(self.name))
        curl = ffi.new('char[]', to_bytes(url))
        err = C.git_submodule_set_url(crepo, cname, curl)
        check_error(err)

    @property
    def branch(self):
        """Branch that is to be tracked by the submodule."""
        branch = C.git_submodule_branch(self._subm)
        return ffi.string(branch).decode('utf-8')

    @property
    def head_id(self) -> Oid | None:
        """
        The submodule's HEAD commit id (as recorded in the superproject's
        current HEAD tree).
        Returns None if the superproject's HEAD doesn't contain the submodule.
        """

        head = C.git_submodule_head_id(self._subm)
        if head == ffi.NULL:
            return None
        return Oid(raw=bytes(ffi.buffer(head.id)[:]))


class SubmoduleCollection:
    """Collection of submodules in a repository."""

    def __init__(self, repository: BaseRepository):
        self._repository = repository

    def __getitem__(self, name: str | Path) -> Submodule:
        """
        Look up submodule information by name or path.
        Raises KeyError if there is no such submodule.
        """
        csub = ffi.new('git_submodule **')
        cpath = ffi.new('char[]', to_bytes(name))

        err = C.git_submodule_lookup(csub, self._repository._repo, cpath)
        check_error(err)
        return Submodule._from_c(self._repository, csub[0])

    def __contains__(self, name: str) -> bool:
        return self.get(name) is not None

    def __iter__(self) -> Iterator[Submodule]:
        for s in self._repository.listall_submodules():
            yield self[s]

    def get(self, name: str) -> Submodule | None:
        """
        Look up submodule information by name or path.
        Unlike __getitem__, this returns None if the submodule is not found.
        """
        try:
            return self[name]
        except KeyError:
            return None

    def add(
        self,
        url: str,
        path: str,
        link: bool = True,
        callbacks: Optional[RemoteCallbacks] = None,
        depth: int = 0,
    ) -> Submodule:
        """
        Add a submodule to the index.
        The submodule is automatically cloned.

        Returns: the submodule that was added.

        Parameters:

        url
            The URL of the submodule.

        path
            The path within the parent repository to add the submodule

        link
            Should workdir contain a gitlink to the repo in `.git/modules` vs. repo directly in workdir.

        callbacks
            Optional RemoteCallbacks to clone the submodule.

        depth
            Number of commits to fetch.
            The default is 0 (full commit history).
        """
        csub = ffi.new('git_submodule **')
        curl = ffi.new('char[]', to_bytes(url))
        cpath = ffi.new('char[]', to_bytes(path))
        gitlink = 1 if link else 0

        err = C.git_submodule_add_setup(
            csub, self._repository._repo, curl, cpath, gitlink
        )
        check_error(err)

        submodule_instance = Submodule._from_c(self._repository, csub[0])

        # Prepare options
        opts = ffi.new('git_submodule_update_options *')
        C.git_submodule_update_options_init(
            opts, C.GIT_SUBMODULE_UPDATE_OPTIONS_VERSION
        )
        opts.fetch_opts.depth = depth

        with git_fetch_options(callbacks, opts=opts.fetch_opts) as payload:
            crepo = ffi.new('git_repository **')
            err = C.git_submodule_clone(crepo, submodule_instance._subm, opts)
            payload.check_error(err)

        # Clean up submodule repository
        from .repository import Repository

        Repository._from_c(crepo[0], True)

        err = C.git_submodule_add_finalize(submodule_instance._subm)
        check_error(err)
        return submodule_instance

    def init(
        self, submodules: Optional[Iterable[str]] = None, overwrite: bool = False
    ) -> None:
        """
        Initialize submodules in the repository. Just like "git submodule init",
        this copies information about the submodules into ".git/config".

        Parameters:

        submodules
            Optional list of submodule paths or names to initialize.
            Default argument initializes all submodules.

        overwrite
            Flag indicating if initialization should overwrite submodule entries.
        """
        if submodules is None:
            submodules = self._repository.listall_submodules()

        instances = [self[s] for s in submodules]

        for submodule in instances:
            submodule.init(overwrite)

    def update(
        self,
        submodules: Optional[Iterable[str]] = None,
        init: bool = False,
        callbacks: Optional[RemoteCallbacks] = None,
        depth: int = 0,
    ) -> None:
        """
        Update submodules. This will clone a missing submodule and checkout
        the subrepository to the commit specified in the index of the
        containing repository. If the submodule repository doesn't contain the
        target commit (e.g. because fetchRecurseSubmodules isn't set), then the
        submodule is fetched using the fetch options supplied in options.

        Parameters:

        submodules
            Optional list of submodule paths or names. If you omit this parameter
            or pass None, all submodules will be updated.

        init
            If the submodule is not initialized, setting this flag to True will
            initialize the submodule before updating. Otherwise, this will raise
            an error if attempting to update an uninitialized repository.

        callbacks
            Optional RemoteCallbacks to clone or fetch the submodule.

        depth
            Number of commits to fetch.
            The default is 0 (full commit history).
        """
        if submodules is None:
            submodules = self._repository.listall_submodules()

        instances = [self[s] for s in submodules]

        for submodule in instances:
            submodule.update(init, callbacks, depth)

    def status(
        self, name: str, ignore: SubmoduleIgnore = SubmoduleIgnore.UNSPECIFIED
    ) -> SubmoduleStatus:
        """
        Get the status of a submodule.

        Returns: A combination of SubmoduleStatus flags.

        Parameters:

        name
            Submodule name or path.

        ignore
            A SubmoduleIgnore value indicating how deeply to examine the working directory.
        """
        cstatus = ffi.new('unsigned int *')
        err = C.git_submodule_status(
            cstatus, self._repository._repo, to_bytes(name), ignore
        )
        check_error(err)
        return SubmoduleStatus(cstatus[0])

    def cache_all(self) -> None:
        """
        Load and cache all submodules in the repository.

        Because the `.gitmodules` file is unstructured, loading submodules is an
        O(N) operation.  Any operation that requires accessing all submodules is O(N^2)
        in the number of submodules, if it has to look each one up individually.
        This function loads all submodules and caches them so that subsequent
        submodule lookups by name are O(1).
        """
        err = C.git_repository_submodule_cache_all(self._repository._repo)
        check_error(err)

    def cache_clear(self) -> None:
        """
        Clear the submodule cache populated by `submodule_cache_all`.
        If there is no cache, do nothing.

        The cache incorporates data from the repository's configuration, as well
        as the state of the working tree, the index, and HEAD. So any time any
        of these has changed, the cache might become invalid.
        """
        err = C.git_repository_submodule_cache_clear(self._repository._repo)
        check_error(err)



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

# Import from pygit2
from ._pygit2 import GitError
from .ffi import C, ffi

__all__ = ['GitError']

value_errors = set([C.GIT_EEXISTS, C.GIT_EINVALIDSPEC, C.GIT_EAMBIGUOUS])


def check_error(err: int, io: bool = False) -> None:
    if err >= 0:
        return

    # These are special error codes, they should never reach here
    test = err != C.GIT_EUSER and err != C.GIT_PASSTHROUGH
    assert test, f'Unexpected error code {err}'

    # Error message
    giterr = C.git_error_last()
    if giterr != ffi.NULL:
        message = ffi.string(giterr.message).decode('utf8', errors='surrogateescape')
    else:
        message = f'err {err} (no message provided)'

    # Translate to Python errors
    if err in value_errors:
        raise ValueError(message)

    if err == C.GIT_ENOTFOUND:
        if io:
            raise IOError(message)

        raise KeyError(message)

    if err == C.GIT_EINVALIDSPEC:
        raise ValueError(message)

    if err == C.GIT_ITEROVER:
        raise StopIteration()

    # Generic Git error
    raise GitError(message)


# Indicate that we want libgit2 to pretend a function was not set
class Passthrough(Exception):
    def __init__(self) -> None:
        super().__init__('The function asked for pass-through')



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

"""
In this module we keep everything concerning callback. This is how it works,
with an example:

1. The pygit2 API calls libgit2, it passes a payload object
   e.g. Remote.fetch calls git_remote_fetch

2. libgit2 calls Python callbacks
   e.g. git_remote_fetch calls _transfer_progress_cb

3. Optionally, the Python callback may proxy to a user defined function
   e.g. _transfer_progress_cb calls RemoteCallbacks.transfer_progress

4. The user defined function may return something on success, or raise an
   exception on error, or raise the special Passthrough exception.

5. The callback may return in 3 different ways to libgit2:

   - Returns GIT_OK on success.
   - Returns GIT_PASSTHROUGH if the user defined function raised Passthrough,
     this tells libgit2 to act as if this callback didn't exist in the first
     place.
   - Returns GIT_EUSER if another exception was raised, and keeps the exception
     in the payload to be re-raised later.

6. libgit2 returns to the pygit2 API, with an error code
   e.g. git_remote_fetch returns to Remote.fetch

7. The pygit2 API will:

   - Return something on success.
   - Raise the original exception if libgit2 returns GIT_EUSER
   - Raise another exception if libgit2 returns another error code

The payload object is passed all the way, so pygit2 API can send information to
the inner user defined function, and this can send back results to the pygit2
API.
"""

# Standard Library
from collections.abc import Callable, Generator
from contextlib import contextmanager
from functools import wraps
from typing import TYPE_CHECKING, Optional, ParamSpec, TypeVar

# pygit2
from ._pygit2 import DiffFile, Oid
from .credentials import Keypair, Username, UserPass
from .enums import CheckoutNotify, CheckoutStrategy, CredentialType, StashApplyProgress
from .errors import Passthrough, check_error
from .ffi import C, ffi
from .utils import StrArray, maybe_string, ptr_to_bytes, to_bytes

_Credentials = Username | UserPass | Keypair

if TYPE_CHECKING:
    from pygit2._libgit2.ffi import GitProxyOptionsC

    from ._pygit2 import CloneOptions, PushOptions
    from .remotes import PushUpdate, TransferProgress
#
# The payload is the way to pass information from the pygit2 API, through
# libgit2, to the Python callbacks. And back.
#


class Payload:
    repository: Callable | None
    remote: Callable | None
    clone_options: 'CloneOptions'

    def __init__(self, **kw: object) -> None:
        for key, value in kw.items():
            setattr(self, key, value)
        self._stored_exception = None

    def check_error(self, error_code: int) -> None:
        if error_code == C.GIT_EUSER:
            assert self._stored_exception is not None
            raise self._stored_exception
        elif self._stored_exception is not None:
            # A callback mapped to a C function returning void
            # might still have raised an exception.
            raise self._stored_exception

        check_error(error_code)


class RemoteCallbacks(Payload):
    """Base class for pygit2 remote callbacks.

    Inherit from this class and override the callbacks which you want to use
    in your class, which you can then pass to the network operations.

    For the credentials, you can either subclass and override the 'credentials'
    method, or if it's a constant value, pass the value to the constructor,
    e.g. RemoteCallbacks(credentials=credentials).

    You can as well pass the certificate the same way, for example:
    RemoteCallbacks(certificate=certificate).
    """

    push_options: 'PushOptions'

    def __init__(
        self,
        credentials: _Credentials | None = None,
        certificate_check: Callable[[None, bool, bytes], bool] | None = None,
    ) -> None:
        super().__init__()
        if credentials is not None:
            self.credentials = credentials  # type: ignore[method-assign, assignment]
        if certificate_check is not None:
            self.certificate_check = certificate_check  # type: ignore[method-assign, assignment]

    def sideband_progress(self, string: str) -> None:
        """
        Progress output callback.  Override this function with your own
        progress reporting function

        Parameters:

        string : str
            Progress output from the remote.
        """

    def credentials(
        self,
        url: str,
        username_from_url: str | None,
        allowed_types: CredentialType,
    ) -> _Credentials:
        """
        Credentials callback.  If the remote server requires authentication,
        this function will be called and its return value used for
        authentication. Override it if you want to be able to perform
        authentication.

        Returns: credential

        Parameters:

        url : str
            The url of the remote.

        username_from_url : str or None
            Username extracted from the url, if any.

        allowed_types : CredentialType
            A combination of CredentialType bitflags representing the
            credential types supported by the remote.
        """
        raise Passthrough

    def certificate_check(self, certificate: None, valid: bool, host: bytes) -> bool:
        """
        Certificate callback. Override with your own function to determine
        whether to accept the server's certificate.

        Returns: True to connect, False to abort.

        Parameters:

        certificate : None
            The certificate. It is currently always None while we figure out
            how to represent it cross-platform.

        valid : bool
            Whether the TLS/SSH library thinks the certificate is valid.

        host : str
            The hostname we want to connect to.
        """

        raise Passthrough

    def push_negotiation(self, updates: list['PushUpdate']) -> None:
        """
        During a push, called once between the negotiation step and the upload.
        Provides information about what updates will be performed.

        Override with your own function to check the pending updates
        and possibly reject them (by raising an exception).
        """

    def transfer_progress(self, stats: 'TransferProgress') -> None:
        """
        During the download of new data, this will be regularly called with
        the indexer's progress.

        Override with your own function to report transfer progress.

        Parameters:

        stats : TransferProgress
            The progress up to now.
        """

    def push_transfer_progress(
        self, objects_pushed: int, total_objects: int, bytes_pushed: int
    ) -> None:
        """
        During the upload portion of a push, this will be regularly called
        with progress information.

        Be aware that this is called inline with pack building operations,
        so performance may be affected.

        Override with your own function to report push transfer progress.
        """

    def update_tips(self, refname: str, old: Oid, new: Oid) -> None:
        """
        Update tips callback. Override with your own function to report
        reference updates.

        Parameters:

        refname : str
            The name of the reference that's being updated.

        old : Oid
            The reference's old value.

        new : Oid
            The reference's new value.
        """

    def push_update_reference(self, refname: str, message: str) -> None:
        """
        Push update reference callback. Override with your own function to
        report the remote's acceptance or rejection of reference updates.

        refname : str
            The name of the reference (on the remote).

        message : str
            Rejection message from the remote. If None, the update was accepted.
        """


class CheckoutCallbacks(Payload):
    """Base class for pygit2 checkout callbacks.

    Inherit from this class and override the callbacks that you want to use
    in your class, which you can then pass to checkout operations.
    """

    def __init__(self) -> None:
        super().__init__()

    def checkout_notify_flags(self) -> CheckoutNotify:
        """
        Returns a bit mask of the notifications to receive from a checkout
        (a combination of enums.CheckoutNotify constants).

        By default, if you override `checkout_notify`, all notifications will
        be enabled. You can fine tune the notification types to enable by
        overriding `checkout_notify_flags`.

        Please note that the flags are only sampled once when checkout begins.
        You cannot change the flags while a checkout is in progress.
        """
        if type(self).checkout_notify == CheckoutCallbacks.checkout_notify:
            # If the user hasn't overridden the notify function,
            # filter out all notifications.
            return CheckoutNotify.NONE
        else:
            # If the user provides their own notify function,
            # enable all notifications by default.
            return CheckoutNotify.ALL

    def checkout_notify(
        self,
        why: CheckoutNotify,
        path: str,
        baseline: Optional[DiffFile],
        target: Optional[DiffFile],
        workdir: Optional[DiffFile],
    ) -> None:
        """
        Checkout will invoke an optional notification callback for
        certain cases - you pick which ones via `checkout_notify_flags`.

        Raising an exception from this callback will cancel the checkout.
        The exception will be propagated back and raised by the
        Repository.checkout_... call.

        Notification callbacks are made prior to modifying any files on disk,
        so canceling on any notification will still happen prior to any files
        being modified.
        """
        pass

    def checkout_progress(
        self, path: str, completed_steps: int, total_steps: int
    ) -> None:
        """
        Optional callback to notify the consumer of checkout progress.
        """
        pass


class StashApplyCallbacks(CheckoutCallbacks):
    """Base class for pygit2 stash apply callbacks.

    Inherit from this class and override the callbacks that you want to use
    in your class, which you can then pass to stash apply or pop operations.
    """

    def stash_apply_progress(self, progress: StashApplyProgress) -> None:
        """
        Stash application progress notification function.

        `progress` is a StashApplyProgress constant.

        Raising an exception from this callback will abort the stash
        application.
        """
        pass


#
# The context managers below wrap the calls to libgit2 functions, which them in
# turn call to callbacks defined later in this module. These context managers
# are used in the pygit2 API, see for instance remote.py
#


@contextmanager
def git_clone_options(payload, opts=None):
    if opts is None:
        opts = ffi.new('git_clone_options *')
        C.git_clone_options_init(opts, C.GIT_CLONE_OPTIONS_VERSION)

    handle = ffi.new_handle(payload)

    # Plug callbacks
    if payload.repository:
        opts.repository_cb = C._repository_create_cb
        opts.repository_cb_payload = handle
    if payload.remote:
        opts.remote_cb = C._remote_create_cb
        opts.remote_cb_payload = handle

    # Give back control
    payload._stored_exception = None
    payload.clone_options = opts
    yield payload


@contextmanager
def git_fetch_options(payload, opts=None):
    if payload is None:
        payload = RemoteCallbacks()

    if opts is None:
        opts = ffi.new('git_fetch_options *')
        C.git_fetch_options_init(opts, C.GIT_FETCH_OPTIONS_VERSION)

    # Plug callbacks
    opts.callbacks.sideband_progress = C._sideband_progress_cb
    opts.callbacks.transfer_progress = C._transfer_progress_cb
    opts.callbacks.update_tips = C._update_tips_cb
    opts.callbacks.credentials = C._credentials_cb
    opts.callbacks.certificate_check = C._certificate_check_cb
    # Payload
    handle = ffi.new_handle(payload)
    opts.callbacks.payload = handle

    # Give back control
    payload.fetch_options = opts
    payload._stored_exception = None
    yield payload


@contextmanager
def git_proxy_options(
    payload: object,
    opts: Optional['GitProxyOptionsC'] = None,
    proxy: None | bool | str = None,
) -> Generator['GitProxyOptionsC', None, None]:
    if opts is None:
        opts = ffi.new('git_proxy_options *')
        C.git_proxy_options_init(opts, C.GIT_PROXY_OPTIONS_VERSION)
    if proxy is None:
        opts.type = C.GIT_PROXY_NONE
    elif proxy is True:
        opts.type = C.GIT_PROXY_AUTO
    elif type(proxy) is str:
        opts.type = C.GIT_PROXY_SPECIFIED
        # Keep url in memory, otherwise memory is freed and bad things happen
        payload.__proxy_url = ffi.new('char[]', to_bytes(proxy))  # type: ignore[attr-defined]
        opts.url = payload.__proxy_url  # type: ignore[attr-defined]
    else:
        raise TypeError('Proxy must be None, True, or a string')
    yield opts


@contextmanager
def git_push_options(payload, opts=None):
    if payload is None:
        payload = RemoteCallbacks()

    opts = ffi.new('git_push_options *')
    C.git_push_options_init(opts, C.GIT_PUSH_OPTIONS_VERSION)

    # Plug callbacks
    opts.callbacks.sideband_progress = C._sideband_progress_cb
    opts.callbacks.transfer_progress = C._transfer_progress_cb
    opts.callbacks.update_tips = C._update_tips_cb
    opts.callbacks.credentials = C._credentials_cb
    opts.callbacks.certificate_check = C._certificate_check_cb
    opts.callbacks.push_update_reference = C._push_update_reference_cb
    opts.callbacks.push_negotiation = C._push_negotiation_cb
    # Per libgit2 sources, push_transfer_progress may incur a performance hit.
    # So, set it only if the user has overridden the no-op stub.
    if (
        type(payload).push_transfer_progress
        is not RemoteCallbacks.push_transfer_progress
    ):
        opts.callbacks.push_transfer_progress = C._push_transfer_progress_cb
    # Payload
    handle = ffi.new_handle(payload)
    opts.callbacks.payload = handle

    # Give back control
    payload.push_options = opts
    payload._stored_exception = None
    yield payload


@contextmanager
def git_remote_callbacks(payload):
    if payload is None:
        payload = RemoteCallbacks()

    cdata = ffi.new('git_remote_callbacks *')
    C.git_remote_init_callbacks(cdata, C.GIT_REMOTE_CALLBACKS_VERSION)

    # Plug callbacks
    cdata.credentials = C._credentials_cb
    cdata.update_tips = C._update_tips_cb
    cdata.certificate_check = C._certificate_check_cb
    # Payload
    handle = ffi.new_handle(payload)
    cdata.payload = handle

    # Give back control
    payload._stored_exception = None
    payload.remote_callbacks = cdata
    yield payload


#
# C callbacks
#
# These functions are called by libgit2. They cannot raise exceptions, since
# they return to libgit2, they can only send back error codes.
#
# They cannot be overridden, but sometimes the only thing these functions do is
# to proxy the call to a user defined function. If user defined functions
# raises an exception, the callback must store it somewhere and return
# GIT_EUSER to libgit2, then the outer Python code will be able to reraise the
# exception.
#

P = ParamSpec('P')
T = TypeVar('T')


def libgit2_callback(f: Callable[P, T]) -> Callable[P, T]:
    @wraps(f)
    def wrapper(*args):
        data = ffi.from_handle(args[-1])
        args = args[:-1] + (data,)
        try:
            return f(*args)
        except Passthrough:
            # A user defined callback can raise Passthrough to decline to act;
            # then libgit2 will behave as if there was no callback set in the
            # first place.
            return C.GIT_PASSTHROUGH
        except BaseException as e:
            # Keep the exception to be re-raised later, and inform libgit2 that
            # the user defined callback has failed.
            data._stored_exception = e
            return C.GIT_EUSER

    return ffi.def_extern()(wrapper)  # type: ignore[attr-defined]


def libgit2_callback_void(f: Callable[P, T]) -> Callable[P, T]:
    @wraps(f)
    def wrapper(*args):
        data = ffi.from_handle(args[-1])
        args = args[:-1] + (data,)
        try:
            f(*args)
        except Passthrough:
            # A user defined callback can raise Passthrough to decline to act;
            # then libgit2 will behave as if there was no callback set in the
            # first place.
            pass  # Function returns void
        except BaseException as e:
            # Keep the exception to be re-raised later
            data._stored_exception = e
            pass  # Function returns void, so we can't do much here.

    return ffi.def_extern()(wrapper)  # type: ignore[attr-defined]


@libgit2_callback
def _certificate_check_cb(cert_i, valid, host, data):
    # We want to simulate what should happen if libgit2 supported pass-through
    # for this callback. For SSH, 'valid' is always False, because it doesn't
    # look at known_hosts, but we do want to let it through in order to do what
    # libgit2 would if the callback were not set.
    try:
        is_ssh = cert_i.cert_type == C.GIT_CERT_HOSTKEY_LIBSSH2

        # python's parsing is deep in the libraries and assumes an OpenSSL-owned cert
        val = data.certificate_check(None, bool(valid), ffi.string(host))
        if not val:
            return C.GIT_ECERTIFICATE
    except Passthrough:
        if is_ssh:
            return 0
        elif valid:
            return 0
        else:
            return C.GIT_ECERTIFICATE

    return 0


@libgit2_callback
def _credentials_cb(cred_out, url, username, allowed, data):
    credentials = getattr(data, 'credentials', None)
    if not credentials:
        return 0

    # convert int flags to enum before forwarding to user code
    allowed = CredentialType(allowed)

    ccred = get_credentials(credentials, url, username, allowed)
    cred_out[0] = ccred[0]
    return 0


@libgit2_callback
def _push_negotiation_cb(updates, num_updates, data):
    from .remotes import PushUpdate

    push_negotiation = getattr(data, 'push_negotiation', None)
    if not push_negotiation:
        return 0

    py_updates = [PushUpdate(updates[i]) for i in range(num_updates)]
    push_negotiation(py_updates)
    return 0


@libgit2_callback
def _push_update_reference_cb(ref, msg, data):
    push_update_reference = getattr(data, 'push_update_reference', None)
    if not push_update_reference:
        return 0

    refname = maybe_string(ref)
    message = maybe_string(msg)
    push_update_reference(refname, message)
    return 0


@libgit2_callback
def _remote_create_cb(remote_out, repo, name, url, data):
    from .repository import Repository

    remote = data.remote(
        Repository._from_c(repo, False), ffi.string(name), ffi.string(url)
    )
    remote_out[0] = remote._remote
    # we no longer own the C object
    remote._remote = ffi.NULL

    return 0


@libgit2_callback
def _repository_create_cb(repo_out, path, bare, data):
    repository = data.repository(ffi.string(path), bare != 0)
    # we no longer own the C object
    repository._disown()
    repo_out[0] = repository._repo

    return 0


@libgit2_callback
def _sideband_progress_cb(string, length, data):
    sideband_progress = getattr(data, 'sideband_progress', None)
    if not sideband_progress:
        return 0

    s = ffi.string(string, length).decode('utf-8')
    sideband_progress(s)
    return 0


@libgit2_callback
def _transfer_progress_cb(stats_ptr, data):
    from .remotes import TransferProgress

    transfer_progress = getattr(data, 'transfer_progress', None)
    if not transfer_progress:
        return 0

    transfer_progress(TransferProgress(stats_ptr))
    return 0


@libgit2_callback
def _push_transfer_progress_cb(current, total, bytes_pushed, payload):
    push_transfer_progress = getattr(payload, 'push_transfer_progress', None)
    if not push_transfer_progress:
        return 0

    push_transfer_progress(current, total, bytes_pushed)
    return 0


@libgit2_callback
def _update_tips_cb(refname, a, b, data):
    update_tips = getattr(data, 'update_tips', None)
    if not update_tips:
        return 0

    s = maybe_string(refname)
    a = Oid(raw=bytes(ffi.buffer(a)[:]))
    b = Oid(raw=bytes(ffi.buffer(b)[:]))
    update_tips(s, a, b)
    return 0


#
# Other functions, used above.
#


def get_credentials(fn, url, username, allowed):
    """Call fn and return the credentials object."""
    url_str = maybe_string(url)
    username_str = maybe_string(username)

    creds = fn(url_str, username_str, allowed)

    credential_type = getattr(creds, 'credential_type', None)
    credential_tuple = getattr(creds, 'credential_tuple', None)
    if not credential_type or not credential_tuple:
        raise TypeError('credential does not implement interface')

    cred_type = credential_type

    if not (allowed & cred_type):
        raise TypeError('invalid credential type')

    ccred = ffi.new('git_credential **')
    if cred_type == CredentialType.USERPASS_PLAINTEXT:
        name, passwd = credential_tuple
        err = C.git_credential_userpass_plaintext_new(
            ccred, to_bytes(name), to_bytes(passwd)
        )

    elif cred_type == CredentialType.SSH_KEY:
        name, pubkey, privkey, passphrase = credential_tuple
        name = to_bytes(name)
        if pubkey is None and privkey is None:
            err = C.git_credential_ssh_key_from_agent(ccred, name)
        else:
            err = C.git_credential_ssh_key_new(
                ccred, name, to_bytes(pubkey), to_bytes(privkey), to_bytes(passphrase)
            )

    elif cred_type == CredentialType.USERNAME:
        (name,) = credential_tuple
        err = C.git_credential_username_new(ccred, to_bytes(name))

    elif cred_type == CredentialType.SSH_MEMORY:
        name, pubkey, privkey, passphrase = credential_tuple
        if pubkey is None and privkey is None:
            raise TypeError('SSH keys from memory are empty')
        err = C.git_credential_ssh_key_memory_new(
            ccred,
            to_bytes(name),
            to_bytes(pubkey),
            to_bytes(privkey),
            to_bytes(passphrase),
        )
    else:
        raise TypeError('unsupported credential type')

    check_error(err)

    return ccred


#
# Checkout callbacks
#


@libgit2_callback
def _checkout_notify_cb(
    why, path_cstr, baseline, target, workdir, data: CheckoutCallbacks
):
    pypath = maybe_string(path_cstr)
    pybaseline = DiffFile.from_c(ptr_to_bytes(baseline))
    pytarget = DiffFile.from_c(ptr_to_bytes(target))
    pyworkdir = DiffFile.from_c(ptr_to_bytes(workdir))

    try:
        data.checkout_notify(why, pypath, pybaseline, pytarget, pyworkdir)  # type: ignore[arg-type]
    except Passthrough:
        # Unlike most other operations with optional callbacks, checkout
        # doesn't support the GIT_PASSTHROUGH return code, so we must bypass
        # libgit2_callback's error handling and return 0 explicitly here.
        pass

    # If the user's callback has raised any other exception type,
    # it's caught by the libgit2_callback decorator by now.
    # So, return success code to libgit2.
    return 0


@libgit2_callback_void
def _checkout_progress_cb(path, completed_steps, total_steps, data: CheckoutCallbacks):
    data.checkout_progress(maybe_string(path), completed_steps, total_steps)  # type: ignore[arg-type]


def _git_checkout_options(
    callbacks=None,
    strategy=None,
    directory=None,
    paths=None,
    c_checkout_options_ptr=None,
):
    if callbacks is None:
        payload = CheckoutCallbacks()
    else:
        payload = callbacks

    # Get handle to payload
    handle = ffi.new_handle(payload)

    # Create the options struct to pass
    if not c_checkout_options_ptr:
        opts = ffi.new('git_checkout_options *')
    else:
        opts = c_checkout_options_ptr
    check_error(C.git_checkout_options_init(opts, 1))

    # References we need to keep to strings and so forth
    refs = [handle]

    # pygit2's default is SAFE | RECREATE_MISSING
    if strategy is None:
        strategy = CheckoutStrategy.SAFE | CheckoutStrategy.RECREATE_MISSING
    opts.checkout_strategy = int(strategy)

    if directory:
        target_dir = ffi.new('char[]', to_bytes(directory))
        refs.append(target_dir)
        opts.target_directory = target_dir

    if paths:
        strarray = StrArray(paths)
        refs.append(strarray)
        opts.paths = strarray.ptr[0]

    # If we want to receive any notifications, set up notify_cb in the options
    notify_flags = payload.checkout_notify_flags()
    if notify_flags != CheckoutNotify.NONE:
        opts.notify_cb = C._checkout_notify_cb
        opts.notify_flags = int(notify_flags)
        opts.notify_payload = handle

    # Set up progress callback if the user has provided their own
    if type(payload).checkout_progress != CheckoutCallbacks.checkout_progress:
        opts.progress_cb = C._checkout_progress_cb
        opts.progress_payload = handle

    # Give back control
    payload.checkout_options = opts
    payload._ffi_handle = handle
    payload._refs = refs
    payload._stored_exception = None
    return payload


@contextmanager
def git_checkout_options(callbacks=None, strategy=None, directory=None, paths=None):
    yield _git_checkout_options(
        callbacks=callbacks, strategy=strategy, directory=directory, paths=paths
    )


#
# Stash callbacks
#


@libgit2_callback
def _stash_apply_progress_cb(progress: StashApplyProgress, data: StashApplyCallbacks):
    try:
        data.stash_apply_progress(progress)
    except Passthrough:
        # Unlike most other operations with optional callbacks, stash apply
        # doesn't support the GIT_PASSTHROUGH return code, so we must bypass
        # libgit2_callback's error handling and return 0 explicitly here.
        pass

    # If the user's callback has raised any other exception type,
    # it's caught by the libgit2_callback decorator by now.
    # So, return success code to libgit2.
    return 0


@contextmanager
def git_stash_apply_options(
    callbacks=None, reinstate_index=False, strategy=None, directory=None, paths=None
):
    if callbacks is None:
        callbacks = StashApplyCallbacks()

    # Set up stash options
    stash_apply_options = ffi.new('git_stash_apply_options *')
    check_error(C.git_stash_apply_options_init(stash_apply_options, 1))

    flags = reinstate_index * C.GIT_STASH_APPLY_REINSTATE_INDEX
    stash_apply_options.flags = flags

    # Now set up checkout options
    c_checkout_options_ptr = ffi.addressof(stash_apply_options.checkout_options)
    payload = _git_checkout_options(
        callbacks=callbacks,
        strategy=strategy,
        directory=directory,
        paths=paths,
        c_checkout_options_ptr=c_checkout_options_ptr,
    )
    assert payload == callbacks
    assert payload.checkout_options == c_checkout_options_ptr

    # Set up stash progress callback if the user has provided their own
    if type(callbacks).stash_apply_progress != StashApplyCallbacks.stash_apply_progress:
        stash_apply_options.progress_cb = C._stash_apply_progress_cb
        stash_apply_options.progress_payload = payload._ffi_handle

    # Give back control
    payload.stash_apply_options = stash_apply_options
    yield payload



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

import contextlib
import os
from collections.abc import Generator, Iterator, Sequence
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Generic,
    Optional,
    Protocol,
    TypeVar,
    Union,
    overload,
)

# Import from pygit2
from .ffi import C, ffi

if TYPE_CHECKING:
    from ._libgit2.ffi import ArrayC, GitStrrayC, char, char_pointer


def maybe_string(ptr: 'char_pointer | None') -> str | None:
    if not ptr:
        return None

    return ffi.string(ptr).decode('utf8', errors='surrogateescape')


@overload
def to_bytes(
    s: str | bytes | os.PathLike[str] | os.PathLike[bytes],
    encoding: str = 'utf-8',
    errors: str = 'strict',
) -> bytes: ...
@overload
def to_bytes(
    s: Union['ffi.NULL_TYPE', None],
    encoding: str = 'utf-8',
    errors: str = 'strict',
) -> Union['ffi.NULL_TYPE']: ...
def to_bytes(
    s: Union[str, bytes, 'ffi.NULL_TYPE', os.PathLike[str], os.PathLike[bytes], None],
    encoding: str = 'utf-8',
    errors: str = 'strict',
) -> Union[bytes, 'ffi.NULL_TYPE']:
    if s == ffi.NULL or s is None:
        return ffi.NULL

    if hasattr(s, '__fspath__'):
        s = os.fspath(s)

    if isinstance(s, bytes):
        return s

    return s.encode(encoding, errors)  # type: ignore[union-attr]


def to_str(s: str | bytes | os.PathLike[str] | os.PathLike[bytes]) -> str:
    if hasattr(s, '__fspath__'):
        s = os.fspath(s)

    if type(s) is str:
        return s

    if type(s) is bytes:
        return s.decode()

    raise TypeError(f'unexpected type "{repr(s)}"')


def ptr_to_bytes(ptr_cdata) -> bytes:
    """
    Convert a pointer coming from C code (<cdata 'some_type *'>)
    to a byte buffer containing the address that the pointer refers to.
    """

    pp = ffi.new('void **', ptr_cdata)
    return bytes(ffi.buffer(pp)[:])


@contextlib.contextmanager
def new_git_strarray() -> Generator['GitStrrayC', None, None]:
    strarray = ffi.new('git_strarray *')
    yield strarray
    C.git_strarray_dispose(strarray)


def strarray_to_strings(arr) -> list[str]:
    """
    Return a list of strings from a git_strarray pointer.

    Free the strings contained in the git_strarry, this means it won't be usable after
    calling this function.
    """
    try:
        return [ffi.string(arr.strings[i]).decode('utf-8') for i in range(arr.count)]
    finally:
        C.git_strarray_dispose(arr)


class StrArray:
    """A git_strarray wrapper

    Use this in order to get a git_strarray* to pass to libgit2 out of a
    list of strings. This has a context manager, which you should use, e.g.

        with StrArray(list_of_strings) as arr:
            C.git_function_that_takes_strarray(arr.ptr)

    To make a pre-existing git_strarray point to the provided list of strings,
    use the context manager's assign_to() method:

        struct = ffi.new('git_strarray *', [ffi.NULL, 0])
        with StrArray(list_of_strings) as arr:
            arr.assign_to(struct)

    The above construct is still subject to FFI scoping rules, i.e. the
    contents of 'struct' only remain valid within the StrArray context.
    """

    __array: 'GitStrrayC | ffi.NULL_TYPE'
    __strings: list['None | ArrayC[char]']
    __arr: 'ArrayC[char_pointer]'

    def __init__(self, lst: None | Sequence[str | os.PathLike[str]]):
        # Allow passing in None as lg2 typically considers them the same as empty
        if lst is None:
            self.__array = ffi.NULL
            return

        if not isinstance(lst, (list, tuple)):
            raise TypeError('Value must be a list')

        strings: list[None | 'ArrayC[char]'] = [None] * len(lst)
        for i in range(len(lst)):
            li = lst[i]
            if not isinstance(li, str) and not hasattr(li, '__fspath__'):
                raise TypeError('Value must be a string or PathLike object')

            strings[i] = ffi.new('char []', to_bytes(li))

        self.__arr = ffi.new('char *[]', strings)
        self.__strings = strings
        self.__array = ffi.new('git_strarray *', [self.__arr, len(strings)])  # type: ignore[call-overload]

    def __enter__(self) -> 'StrArray':
        return self

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        pass

    @property
    def ptr(self) -> 'GitStrrayC | ffi.NULL_TYPE':
        return self.__array

    def assign_to(self, git_strarray: 'GitStrrayC') -> None:
        if self.__array == ffi.NULL:
            git_strarray.strings = ffi.NULL
            git_strarray.count = 0
        else:
            git_strarray.strings = self.__arr
            git_strarray.count = len(self.__strings)


T = TypeVar('T')
U = TypeVar('U', covariant=True)


class SequenceProtocol(Protocol[U]):
    def __len__(self) -> int: ...
    def __getitem__(self, index: int) -> U: ...


class GenericIterator(Generic[T]):
    """Helper to easily implement an iterator.

    The constructor gets a container which must implement __len__ and
    __getitem__
    """

    def __init__(self, container: SequenceProtocol[T]) -> None:
        self.container = container
        self.length = len(container)
        self.idx = 0

    def __iter__(self) -> Iterator[T]:
        return self

    def __next__(self) -> T:
        idx = self.idx
        if idx >= self.length:
            raise StopIteration

        self.idx += 1
        return self.container[idx]



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

from __future__ import annotations

import weakref
from collections.abc import Callable
from typing import TYPE_CHECKING

from ._pygit2 import Blob, FilterSource
from .errors import check_error
from .ffi import C, ffi
from .utils import to_bytes

if TYPE_CHECKING:
    from ._libgit2.ffi import GitFilterListC
    from .repository import BaseRepository


class Filter:
    """
    Base filter class to be used with libgit2 filters.

    Inherit from this class and override the `check()`, `write()` and `close()`
    methods to define a filter which can then be registered via
    `pygit2.filter_register()`.

    A new Filter instance will be instantiated for each stream which needs to
    be filtered. For each stream, filter methods will be called in this order:

        - `check()`
        - `write()` (may be called multiple times)
        - `close()`

    Filtered output data should be written to the next filter in the chain
    during `write()` and `close()` via the `write_next` method. All output data
    must be written to the next filter before returning from `close()`.

    If a filter is dependent on reading the complete input data stream, the
    filter should only write output data in `close()`.
    """

    #: Space-separated string list of attributes to be used in `check()`
    attributes: str = ''

    @classmethod
    def nattrs(cls) -> int:
        return len(cls.attributes.split())

    def check(self, src: FilterSource, attr_values: list[str | None]) -> None:
        """
        Check whether this filter should be applied to the given source.

        `check` will be called once per stream.

        If `Passthrough` is raised, the filter will not be applied.

        Parameters:

        src: The source of the filtered blob.

        attr_values: The values of each attribute for the blob being filtered.
            `attr_values` will be a sorted list containing attributes in the
            order they were defined in ``cls.attributes``.
        """

    def write(
        self, data: bytes, src: FilterSource, write_next: Callable[[bytes], None]
    ) -> None:
        """
        Write input `data` to this filter.

        `write()` may be called multiple times per stream.

        Parameters:

        data: Input data.

        src: The source of the filtered blob.

        write_next: The ``write()`` method of the next filter in the chain.
            Filtered output data should be written to `write_next` whenever it is
            available.
        """
        write_next(data)

    def close(self, write_next: Callable[[bytes], None]) -> None:
        """
        Close this filter.

        `close()` will be called once per stream whenever all writes() to this
        stream have been completed.

        Parameters:
            write_next: The ``write()`` method of the next filter in the chain.
                Any remaining filtered output data must be written to
                `write_next` before returning.
        """


class FilterList:
    _all_filter_lists: set[weakref.ReferenceType[FilterList]] = set()

    _pointer: GitFilterListC

    @classmethod
    def _from_c(cls, ptr: GitFilterListC):
        if ptr == ffi.NULL:
            return None

        fl = cls.__new__(cls)
        fl._pointer = ptr

        # Keep track of this FilterList until it's garbage collected. This lets
        # `filter_unregister` ensure the user isn't trying to delete a filter
        # that's still in use.
        ref = weakref.ref(fl, FilterList._all_filter_lists.remove)
        FilterList._all_filter_lists.add(ref)

        return fl

    @classmethod
    def _is_filter_in_use(cls, name: str) -> bool:
        for ref in cls._all_filter_lists:
            fl = ref()
            if fl is not None and name in fl:
                return True
        return False

    def __contains__(self, name: str) -> bool:
        if not isinstance(name, str):
            raise TypeError('argument must be str')
        c_name = to_bytes(name)
        result = C.git_filter_list_contains(self._pointer, c_name)
        return bool(result)

    def __len__(self) -> int:
        return C.git_filter_list_length(self._pointer)

    def apply_to_buffer(self, data: bytes) -> bytes:
        """
        Apply a filter list to a data buffer.
        Return the filtered contents.
        """
        buf = ffi.new('git_buf *')
        err = C.git_filter_list_apply_to_buffer(buf, self._pointer, data, len(data))
        check_error(err)
        try:
            return ffi.string(buf.ptr)
        finally:
            C.git_buf_dispose(buf)

    def apply_to_file(self, repo: BaseRepository, path: str) -> bytes:
        """
        Apply a filter list to the contents of a file on disk.
        Return the filtered contents.
        """
        buf = ffi.new('git_buf *')
        c_path = to_bytes(path)
        err = C.git_filter_list_apply_to_file(buf, self._pointer, repo._repo, c_path)
        check_error(err)
        try:
            return ffi.string(buf.ptr)
        finally:
            C.git_buf_dispose(buf)

    def apply_to_blob(self, blob: Blob) -> bytes:
        """
        Apply a filter list to a data buffer.
        Return the filtered contents.
        """
        buf = ffi.new('git_buf *')

        c_blob = ffi.new('git_blob **')
        ffi.buffer(c_blob)[:] = blob._pointer[:]

        err = C.git_filter_list_apply_to_blob(buf, self._pointer, c_blob[0])
        check_error(err)
        try:
            return ffi.string(buf.ptr)
        finally:
            C.git_buf_dispose(buf)

    def __del__(self):
        C.git_filter_list_free(self._pointer)



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

from os import PathLike
from typing import TYPE_CHECKING

# Import from pygit2
from .errors import check_error
from .ffi import C, ffi
from .utils import to_bytes

if TYPE_CHECKING:
    from pygit2 import Oid, Repository
    from pygit2.repository import BaseRepository


class PackBuilder:
    def __init__(self, repo: 'Repository | BaseRepository') -> None:
        cpackbuilder = ffi.new('git_packbuilder **')
        err = C.git_packbuilder_new(cpackbuilder, repo._repo)
        check_error(err)

        self._repo = repo
        self._packbuilder = cpackbuilder[0]
        self._cpackbuilder = cpackbuilder

    @property
    def _pointer(self) -> bytes:
        return bytes(ffi.buffer(self._packbuilder)[:])

    def __del__(self) -> None:
        C.git_packbuilder_free(self._packbuilder)

    def __len__(self) -> int:
        return C.git_packbuilder_object_count(self._packbuilder)

    @staticmethod
    def __convert_object_to_oid(oid: 'Oid') -> 'ffi.GitOidC':
        git_oid = ffi.new('git_oid *')
        ffi.buffer(git_oid)[:] = oid.raw[:]
        return git_oid

    def add(self, oid: 'Oid') -> None:
        git_oid = self.__convert_object_to_oid(oid)
        err = C.git_packbuilder_insert(self._packbuilder, git_oid, ffi.NULL)
        check_error(err)

    def add_recur(self, oid: 'Oid') -> None:
        git_oid = self.__convert_object_to_oid(oid)
        err = C.git_packbuilder_insert_recur(self._packbuilder, git_oid, ffi.NULL)
        check_error(err)

    def set_threads(self, n_threads: int) -> int:
        return C.git_packbuilder_set_threads(self._packbuilder, n_threads)

    def write(self, path: str | bytes | PathLike[str] | None = None) -> None:
        path_bytes = ffi.NULL if path is None else to_bytes(path)
        err = C.git_packbuilder_write(
            self._packbuilder, path_bytes, 0, ffi.NULL, ffi.NULL
        )
        check_error(err)

    @property
    def written_objects_count(self) -> int:
        return C.git_packbuilder_written(self._packbuilder)



import io
import threading
import time
from contextlib import AbstractContextManager
from queue import Queue
from typing import Optional

from ._pygit2 import Blob, Oid
from .enums import BlobFilter


class _BlobIO(io.RawIOBase):
    """Low-level wrapper for streaming blob content.

    The underlying libgit2 git_writestream filter chain will be run
    in a separate thread. The GIL will be released while running
    libgit2 filtering.
    """

    def __init__(
        self,
        blob: Blob,
        as_path: Optional[str] = None,
        flags: BlobFilter = BlobFilter.CHECK_FOR_BINARY,
        commit_id: Optional[Oid] = None,
    ):
        super().__init__()
        self._blob = blob
        self._queue: Optional[Queue] = Queue(maxsize=1)
        self._ready = threading.Event()
        self._writer_closed = threading.Event()
        self._chunk: Optional[bytes] = None
        self._thread = threading.Thread(
            target=self._blob._write_to_queue,
            args=(self._queue, self._ready, self._writer_closed),
            kwargs={
                'as_path': as_path,
                'flags': int(flags),
                'commit_id': commit_id,
            },
            daemon=True,
        )
        self._thread.start()

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def isatty(self):
        return False

    def readable(self):
        return True

    def writable(self):
        return False

    def seekable(self):
        return False

    def readinto(self, b, /):
        try:
            while self._chunk is None:
                self._ready.wait()
                if self._queue.empty():
                    if self._writer_closed.is_set():
                        # EOF
                        return 0
                    self._ready.clear()
                    time.sleep(0)
                    continue
                chunk = self._queue.get()
                if chunk:
                    self._chunk = chunk

            if len(self._chunk) <= len(b):
                bytes_written = len(self._chunk)
                b[:bytes_written] = self._chunk
                self._chunk = None
                return bytes_written
            bytes_written = len(b)
            b[:] = self._chunk[:bytes_written]
            self._chunk = self._chunk[bytes_written:]
            return bytes_written
        except KeyboardInterrupt:
            return 0

    def close(self) -> None:
        try:
            self._ready.wait()
            self._writer_closed.wait()
            while self._queue is not None and not self._queue.empty():
                self._queue.get()
            self._thread.join()
        except KeyboardInterrupt:
            pass
        self._queue = None


class BlobIO(io.BufferedReader, AbstractContextManager):
    """Read-only wrapper for streaming blob content.

    Supports reading both raw and filtered blob content.
    Implements io.BufferedReader.

    Example:

        >>> with BlobIO(blob) as f:
        ...     while True:
        ...         # Read blob data in 1KB chunks until EOF is reached
        ...         chunk = f.read(1024)
        ...         if not chunk:
        ...             break

    By default, `BlobIO` will stream the raw contents of the blob, but it
    can also be used to stream filtered content (i.e. to read the content
    after applying filters which would be used when checking out the blob
    to the working directory).

    Example:

        >>> with BlobIO(blob, as_path='my_file.ext') as f:
        ...     # Read the filtered content which would be returned upon
        ...     # running 'git checkout -- my_file.txt'
        ...     filtered_data = f.read()
    """

    def __init__(
        self,
        blob: Blob,
        as_path: Optional[str] = None,
        flags: BlobFilter = BlobFilter.CHECK_FOR_BINARY,
        commit_id: Optional[Oid] = None,
    ):
        """Wrap the specified blob.

        Parameters:
            blob: The blob to wrap.
            as_path: Filter the contents of the blob as if it had the specified
                path. If `as_path` is None, the raw contents of the blob will
                be read.
            flags: A combination of enums.BlobFilter constants
                (only applicable when `as_path` is set).
            commit_id: Commit to load attributes from when
                ATTRIBUTES_FROM_COMMIT is specified in `flags`
                (only applicable when `as_path` is set).
        """
        raw = _BlobIO(blob, as_path=as_path, flags=flags, commit_id=commit_id)
        super().__init__(raw)

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


io.RawIOBase.register(_BlobIO)
io.BufferedIOBase.register(BlobIO)



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

from __future__ import annotations

from collections.abc import Iterator
from typing import TYPE_CHECKING

from pygit2 import Oid

from .enums import ReferenceFilter

# Need BaseRepository for type hints, but don't let it cause a circular dependency
if TYPE_CHECKING:
    from ._pygit2 import Reference
    from .repository import BaseRepository


class References:
    def __init__(self, repository: BaseRepository) -> None:
        self._repository = repository

    def __getitem__(self, name: str) -> 'Reference':
        return self._repository.lookup_reference(name)

    def get(self, key: str) -> 'Reference' | None:
        try:
            return self[key]
        except KeyError:
            return None

    def __iter__(self) -> Iterator[str]:
        iter = self._repository.references_iterator_init()
        while True:
            ref = self._repository.references_iterator_next(iter)
            if ref:
                yield ref.name
            else:
                return

    def iterator(
        self, references_return_type: ReferenceFilter = ReferenceFilter.ALL
    ) -> Iterator['Reference']:
        """Creates a new iterator and fetches references for a given repository.

        Can also filter and pass all refs or only branches or only tags.

        Parameters:

        references_return_type: ReferenceFilter
            Optional specifier to filter references. By default, all references are
            returned.

            The following values are accepted:
            - ReferenceFilter.ALL, fetches all refs, this is the default
            - ReferenceFilter.BRANCHES, fetches only branches
            - ReferenceFilter.TAGS, fetches only tags

        TODO: Add support for filtering by reference types notes and remotes.
        """

        # Enforce ReferenceFilter type - raises ValueError if we're given an invalid value
        references_return_type = ReferenceFilter(references_return_type)

        iter = self._repository.references_iterator_init()
        while True:
            ref = self._repository.references_iterator_next(
                iter, references_return_type
            )
            if ref:
                yield ref
            else:
                return

    def create(self, name: str, target: Oid | str, force: bool = False) -> 'Reference':
        return self._repository.create_reference(name, target, force)

    def delete(self, name: str) -> None:
        self[name].delete()

    def __contains__(self, name: str) -> bool:
        return self.get(name) is not None

    @property
    def objects(self) -> list['Reference']:
        return self._repository.listall_reference_objects()

    def compress(self) -> None:
        return self._repository.compress_references()
